from __future__ import annotations

from contextvars import ContextVar
from dataclasses import dataclass, field
from typing import Any, Callable, Generic, TypeVar, Union, overload
from weakref import WeakKeyDictionary

from typing_extensions import Self, TypeVarTuple, Unpack

from avilla.core.utilles import classproperty

T = TypeVar("T")

_MetadataT1 = TypeVar("_MetadataT1", bound="Metadata")
_MetadataT2 = TypeVar("_MetadataT2", bound="Metadata")

_TVT1 = TypeVarTuple("_TVT1")
_TVT2 = TypeVarTuple("_TVT2")


class _GetItemAgent(Generic[T]):
    __getitem__: T

    def __init__(self, getitem: T):
        self.__getitem__ = getitem


class MetadataMeta(type):
    @overload
    def __rshift__(cls: type[_MetadataT1], other: type[_MetadataT2]) -> MetadataRoute[_MetadataT1, _MetadataT2]:
        ...

    @overload
    def __rshift__(
        cls: type[_MetadataT1], other: MetadataRoute[Unpack[_TVT1]]
    ) -> MetadataRoute[_MetadataT1, Unpack[_TVT1]]:
        ...

    def __rshift__(cls: Any, other: type[Metadata] | MetadataRoute) -> MetadataRoute:
        # sourcery skip: instance-method-first-arg-name
        cells = (cls, other) if isinstance(other, type) else (cls, *other.cells)
        return MetadataRoute(cells)


METACELL_PARAMS_CTX: WeakKeyDictionary[type[Metadata], ContextVar[dict[str, Any] | None]] = WeakKeyDictionary()


@dataclass
class Metadata(metaclass=MetadataMeta):
    route: type[Self] | MetadataRoute = field(init=False)

    def infers(self, route: type[Self] | MetadataRoute):
        self.route = route
        return self

    def __post_init__(self):
        self.route = type(self)

    @classproperty
    @classmethod
    def _param_ctx(cls):
        return METACELL_PARAMS_CTX[cls]

    def __init_subclass__(cls) -> None:
        METACELL_PARAMS_CTX[cls] = ContextVar(f"$MetadataParam${cls.__module__}::{cls.__qualname__}", default=None)
        super().__init_subclass__()

    @classmethod
    def inh(cls: type[_MetadataT1], operator: Callable[[_MetadataT1], T]) -> FieldReference[_MetadataT1, T]:
        return FieldReference(cls, operator)

    @classmethod
    def get_params(cls) -> dict[str, Any] | None:
        return cls._param_ctx.get()

    @classmethod
    def set_params(cls, params: dict[str, Any]):
        target = cls._param_ctx.get()
        if target is None:
            target = {}
            cls._param_ctx.set(target)
        target.update(params)

    @classmethod
    def clear_params(cls) -> None:
        cls._param_ctx.set(None)

    @classmethod
    def has_params(cls) -> bool:
        return cls._param_ctx.get() is not None


class MetadataRoute(Generic[Unpack[_TVT1]]):
    cells: tuple[type[Metadata], ...]

    def __init__(self, cells: tuple[type[Metadata], ...]) -> None:
        self.cells = cells

    @overload
    def __rshift__(
        self: MetadataRoute[Unpack[_TVT1]], other: type[_MetadataT1]
    ) -> MetadataRoute[Unpack[_TVT1], _MetadataT1]:
        ...

    @overload
    def __rshift__(
        self: MetadataRoute[Unpack[_TVT1]], other: MetadataRoute[Unpack[_TVT2]]
    ) -> MetadataRoute[Unpack[_TVT1], Unpack[_TVT2]]:
        ...

    def __rshift__(self, other: type[Metadata] | MetadataRoute) -> MetadataRoute:
        if not isinstance(other, (type, MetadataRoute)):
            raise TypeError(f"{other.__class__} is not allowed.")
        cells = (*self.cells, other) if isinstance(other, type) else (*self.cells, *other.cells)
        return MetadataRoute(cells)

    def __repr__(self) -> str:
        cells_repr = " >> ".join(cell.__name__ for cell in self.cells)
        return f"MetadataRoute[{cells_repr}]"

    def __hash__(self) -> int:
        return hash(self.cells) + hash("MetadataRoute")

    def __eq__(self, o: object) -> bool:
        return isinstance(o, MetadataRoute) and o.cells == self.cells

    def clear_params(self) -> None:
        for cell in self.cells:
            cell.clear_params()

    @property
    def _(self: MetadataRoute[Unpack[tuple[Any, ...]], _MetadataT1]):
        @_GetItemAgent
        def wrapper(operator: Callable[[_MetadataT1], T]) -> FieldReference[_MetadataT1, T]:
            return FieldReference(self, operator)

        return wrapper

    def has_params(self) -> bool:
        return any(cell.has_params() for cell in self.cells)


@dataclass(unsafe_hash=True)
class FieldReference(Generic[_MetadataT1, T]):
    define: type[_MetadataT1] | MetadataRoute[Unpack[tuple[Any, ...]], _MetadataT1]
    operator: Callable[[_MetadataT1], T]


Route = Union[type[Metadata], MetadataRoute]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .perform import BasePerform
    from .staff import Staff


class Gateway:
    registry: dict[type[BasePerform], Staff]

    def __init__(self) -> None:
        self.registry = {}

    def register(self, staff: Staff, *perform_types: type[BasePerform]):
        for perform_type in perform_types:
            self.registry[perform_type] = staff


GLOBAL_GATEWAY = Gateway()



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, Protocol, Union, runtime_checkable

from typing_extensions import ParamSpec, TypeVar

if TYPE_CHECKING:
    from .collector import BaseCollector  # noqa

P = ParamSpec("P")
P1 = ParamSpec("P1")
R = TypeVar("R", infer_variance=True)
T = TypeVar("T")
C = TypeVar("C", bound="BaseCollector")
CQ = TypeVar("CQ", bound="BaseCollector")


class SupportsCollect(Protocol[P, R]):
    def collect(self, collector: Any, *args: P.args, **kwargs: P.kwargs) -> R:  # type: ignore
        ...


R1 = TypeVar("R1", covariant=True)
R2 = TypeVar("R2", covariant=True)


@dataclass(frozen=True, eq=True)
class Twin(Generic[R1, R2]):
    _0: R1
    _1: R2


class ClsIncluded(Protocol[T]):
    cls: type[T]


Cr = TypeVar("Cr", bound="BaseCollector", covariant=True)


@runtime_checkable
class SupportsMerge(Protocol):
    def merge(self, *records: dict):
        ...


class LayoutProtocolProperty(Protocol):
    @property
    def get_artifact_layout(self) -> dict:
        ...


class LayoutProtocolAttr(Protocol):
    get_artifact_layout: dict


LayoutProtocol = Union[LayoutProtocolProperty, LayoutProtocolAttr]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface

from avilla.standard.core.application.event import AvillaLifecycleEvent

if TYPE_CHECKING:
    from avilla.core.account import BaseAccount


@dataclass
class AccountStatusChanged(AvillaLifecycleEvent):
    """指示当前账号 (Account) 状态发生改变的事件"""

    account: "BaseAccount"

    class Dispatcher(BaseDispatcher):
        @classmethod
        async def catch(cls, interface: DispatcherInterface[AccountStatusChanged]):
            from avilla.core.account import BaseAccount

            if issubclass(interface.annotation, BaseAccount):
                return interface.event.account

            return await AvillaLifecycleEvent.Dispatcher.catch(interface)


class AccountAvailable(AccountStatusChanged):
    """指示当前账号处于可用状态."""


class AccountRegistered(AccountAvailable):
    """指示当前账号 (Account) 被注册的事件"""


class AccountUnavailable(AccountStatusChanged):
    """指示当前账号处于不可用状态."""


class AccountUnregistered(AccountUnavailable):
    """指示当前账号 (Account) 被注销的事件"""



import asyncio
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Generic,
    Optional,
    TypeVar,
    Union,
    cast,
    get_args,
    overload,
)

from arclet.alconna import (
    Alconna,
    Arg,
    Args,
    ArgsStub,
    Arparma,
    CommandMeta,
    Duplication,
    Empty,
    Namespace,
    OptionStub,
    SubcommandStub,
    command_manager,
    config,
    output_manager,
)
from arclet.alconna.args import TAValue
from arclet.alconna.argv import Argv, argv_config, set_default_argv_type
from arclet.alconna.builtin import generate_duplication
from arclet.alconna.tools.construct import AlconnaString, alconna_from_format
from creart import it
from graia.amnesia.message import MessageChain
from graia.amnesia.message.element import Text
from graia.broadcast import Broadcast
from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.exectarget import ExecTarget
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from graia.broadcast.typing import T_Dispatcher
from nepattern import DirectPattern
from pygtrie import CharTrie
from tarina.generic import generic_isinstance, generic_issubclass, get_origin
from tarina.string import split_once

from avilla.core import Context, MessageReceived, Notice

T = TypeVar("T")
TCallable = TypeVar("TCallable", bound=Callable[..., Any])


@dataclass
class Match(Generic[T]):
    """
    匹配项，表示参数是否存在于 `all_matched_args` 内

    result (T): 匹配结果

    available (bool): 匹配状态
    """

    result: T
    available: bool


class MessageChainArgv(Argv[MessageChain]):
    @staticmethod
    def generate_token(data: list) -> int:
        return hash("".join(i.__repr__() for i in data))


set_default_argv_type(MessageChainArgv)

argv_config(
    MessageChainArgv,
    filter_out=[],
    checker=lambda x: isinstance(x, MessageChain),
    to_text=lambda x: x.text if x.__class__ is Text else None,
    converter=lambda x: MessageChain(x if isinstance(x, list) else [Text(x)]),
)


def _is_tome(message: MessageChain, context: Context):
    if message.content and isinstance(message[0], Notice):
        notice: Notice = message.get_first(Notice)
        if notice.target.last_value == context.self.last_value:
            return True
    return False


def _remove_tome(message: MessageChain, context: Context):
    if _is_tome(message, context):
        message = MessageChain(message.content.copy())
        message.content.remove(message.get_first(Notice))
        if message.content and isinstance(message.content[0], Text):
            text = message.content[0].text.lstrip()  # type: ignore
            if not text:
                message.content.pop(0)
            else:
                message.content[0] = Text(text)
        return message
    return message


@dataclass
class AlconnaDispatcher(BaseDispatcher):
    cmd: Alconna

    async def catch(self, interface: DispatcherInterface):
        arp = command_manager.get_result(self.cmd)[0]
        default_duplication = generate_duplication(self.cmd)(arp)
        if interface.annotation is Duplication:
            return default_duplication
        if generic_issubclass(Duplication, interface.annotation):
            return interface.annotation(arp)
        if interface.annotation is ArgsStub:
            arg = ArgsStub(self.cmd.args)
            arg.set_result(arp.all_matched_args)
            return arg
        if interface.annotation is OptionStub:
            return default_duplication.option(interface.name)
        if interface.annotation is SubcommandStub:
            return default_duplication.subcommand(interface.name)
        if generic_issubclass(get_origin(interface.annotation), Arparma):
            return arp
        if generic_issubclass(interface.annotation, Alconna):
            return self.cmd
        if interface.annotation is Match:
            r = arp.all_matched_args.get(interface.name, Empty)
            return Match(r, r != Empty)
        if get_origin(interface.annotation) is Match:
            r = arp.all_matched_args.get(interface.name, Empty)
            return Match(r, generic_isinstance(r, get_args(interface.annotation)[0]))
        if interface.name in arp.all_matched_args:
            if generic_isinstance(arp.all_matched_args[interface.name], interface.annotation):
                return arp.all_matched_args[interface.name]
            return


class AvillaCommands:
    __namespace__ = "Avilla"

    def __init__(self, need_tome: bool = False, remove_tome: bool = False):
        self.trie: CharTrie = CharTrie()
        self.broadcast = it(Broadcast)
        self.need_tome = need_tome
        self.remove_tome = remove_tome
        config.namespaces["Avilla"] = Namespace(self.__namespace__)

        @self.broadcast.receiver(MessageReceived)
        async def listener(event: MessageReceived):
            msg = str(event.message.content.exclude(Notice)).lstrip()
            if matches := list(self.trie.prefixes(msg)):
                await asyncio.gather(*(self.execute(*res.value, event) for res in matches if res.value))
                return
            # shortcut
            head, _ = split_once(msg, (" ",))
            for value in self.trie.values():
                try:
                    command_manager.find_shortcut(value[0], head)
                except ValueError:
                    continue
                await self.execute(*value, event)

    @property
    def all_helps(self) -> str:
        return command_manager.all_command_help(namespace=self.__namespace__)

    def get_help(self, command: str) -> str:
        return command_manager.get_command(f"{self.__namespace__}::{command}").get_help()

    async def execute(
        self, command: Alconna, target: ExecTarget, need_tome: bool, remove_tome: bool, event: MessageReceived
    ):
        if (need_tome or self.need_tome) and not _is_tome(event.message.content, event.context):
            return
        with output_manager.capture(command.name) as cap:
            output_manager.set_action(lambda x: x, command.name)
            msg = event.message.content
            if remove_tome or self.remove_tome:
                msg = _remove_tome(msg, event.context)
            try:
                _res = command.parse(msg)
            except Exception as e:
                _res = Arparma(command.path, event.message.content, False, error_info=e)
            may_help_text: Optional[str] = cap.get("output", None)
        if _res.matched:
            await self.broadcast.Executor(target, [event.Dispatcher, AlconnaDispatcher(command)])
            target.oplog.clear()
        elif may_help_text:
            await event.context.scene.send_message(may_help_text)

    def command(
        self,
        command: str,
        help_text: Optional[str] = None,
        need_tome: bool = False,
        remove_tome: bool = False,
        dispatchers: Optional[list[T_Dispatcher]] = None,
        decorators: Optional[list[Decorator]] = None,
    ):
        class Command(AlconnaString):
            def __call__(_cmd_self, func: TCallable) -> TCallable:
                return self.on(_cmd_self.build(), need_tome, remove_tome, dispatchers, decorators)(func)

        return Command(command, help_text)

    @overload
    def on(
        self,
        command: Alconna,
        need_tome: bool = False,
        remove_tome: bool = False,
        dispatchers: Optional[list[T_Dispatcher]] = None,
        decorators: Optional[list[Decorator]] = None,
    ) -> Callable[[TCallable], TCallable]:
        ...

    @overload
    def on(
        self,
        command: str,
        need_tome: bool = False,
        remove_tome: bool = False,
        dispatchers: Optional[list[T_Dispatcher]] = None,
        decorators: Optional[list[Decorator]] = None,
        *,
        args: Optional[dict[str, Union[TAValue, Args, Arg]]] = None,
        meta: Optional[CommandMeta] = None,
    ) -> Callable[[TCallable], TCallable]:
        ...

    def on(
        self,
        command: Union[Alconna, str],
        need_tome: bool = False,
        remove_tome: bool = False,
        dispatchers: Optional[list[T_Dispatcher]] = None,
        decorators: Optional[list[Decorator]] = None,
        *,
        args: Optional[dict[str, Union[TAValue, Args, Arg]]] = None,
        meta: Optional[CommandMeta] = None,
    ) -> Callable[[TCallable], TCallable]:
        def wrapper(func: TCallable) -> TCallable:
            target = ExecTarget(func, dispatchers, decorators)
            if isinstance(command, str):
                mapping = {arg.name: arg.value for arg in Args.from_callable(func)[0]}
                mapping.update(args or {})  # type: ignore
                _command = alconna_from_format(command, mapping, meta, union=False)
                _command.reset_namespace(self.__namespace__)
                key = _command.name + "".join(
                    f" {arg.value.target}" for arg in _command.args if isinstance(arg.value, DirectPattern)
                )
                self.trie[key] = (_command, target, need_tome, remove_tome)
            else:
                if not isinstance(command.command, str):
                    raise TypeError("Command name must be a string.")
                if not command.prefixes:
                    self.trie[command.command] = (command, target, need_tome, remove_tome)
                elif not all(isinstance(i, str) for i in command.prefixes):
                    raise TypeError("Command prefixes must be a list of string.")
                else:
                    for prefix in cast(list[str], command.prefixes):
                        self.trie[prefix + command.command] = (command, target, need_tome, remove_tome)
                command.reset_namespace(self.__namespace__)
            return func

        return wrapper


__all__ = ["AvillaCommands", "Match"]



from __future__ import annotations

from datetime import timedelta

from avilla.core.ryanvk import Capability, Fn, TargetOverload
from avilla.core.selector import Selector


class PrivilegeCapability(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def upgrade(self, target: Selector, dest: str | None = None) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def downgrade(self, target: Selector, dest: str | None = None) -> None:
        ...


class MuteCapability(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def mute(self, target: Selector, duration: timedelta) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def unmute(self, target: Selector) -> None:
        ...


class MuteAllCapability(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def mute_all(self, target: Selector) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def unmute_all(self, target: Selector) -> None:
        ...

    # Fetch => rs.pull(MuteInfo, target=...)


class BanCapability(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def ban(self, target: Selector, *, duration: timedelta | None = None, reason: str | None = None) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def unban(self, target: Selector) -> None:
        ...



from __future__ import annotations

from contextlib import nullcontext
from typing import TYPE_CHECKING, Any, ClassVar

from typing_extensions import Self

from avilla.core._runtime import cx_avilla, cx_context, cx_protocol
from avilla.core.event import AvillaEvent

if TYPE_CHECKING:
    from avilla.core.application import Avilla
    from avilla.core.context import Context


class ProtocolConfig:
    ...


class BaseProtocol:
    avilla: Avilla
    artifacts: ClassVar[dict[Any, Any]]

    def ensure(self, avilla: Avilla) -> Any:
        ...

    def configure(self, config: ProtocolConfig) -> Self:
        ...

    def post_event(self, event: AvillaEvent, context: Context | None = None):
        with cx_avilla.use(self.avilla), cx_protocol.use(self), (
            cx_context.use(context) if context is not None else nullcontext()
        ):
            self.avilla.event_record(event)
            return self.avilla.broadcast.postEvent(event)



from __future__ import annotations

from collections.abc import Callable
from typing import Any, TypedDict, TypeVar, cast, overload

from typing_extensions import ParamSpec, Unpack

from avilla.core._runtime import cx_context
from avilla.core.account import BaseAccount
from avilla.core.metadata import Metadata, MetadataRoute
from avilla.core.platform import Land
from avilla.core.resource import Resource
from avilla.core.ryanvk import Fn
from avilla.core.ryanvk.staff import Staff
from avilla.core.selector import FollowsPredicater, Selectable, Selector
from avilla.core.utilles import classproperty

from ._roles import (
    ContextClientSelector,
    ContextEndpointSelector,
    ContextMedium,
    ContextRequestSelector,
    ContextSceneSelector,
    ContextSelfSelector,
)
from ._selector import ContextSelector

P = ParamSpec("P")
R = TypeVar("R", covariant=True)
_T = TypeVar("_T")
_MetadataT = TypeVar("_MetadataT", bound=Metadata)


class ContextCache(TypedDict):
    meta: dict[Selector, dict[type[Metadata] | MetadataRoute, Metadata]]


class Context:
    account: BaseAccount

    client: ContextClientSelector
    endpoint: ContextEndpointSelector
    scene: ContextSceneSelector
    self: ContextSelfSelector
    mediums: list[ContextMedium]

    cache: ContextCache | dict[str, Any]

    def __init__(
        self,
        account: BaseAccount,
        client: Selector,
        endpoint: Selector,
        scene: Selector,
        selft: Selector,
        mediums: list[Selector] | None = None,
        prelude_metadatas: dict[Selector, dict[type[Metadata] | MetadataRoute, Metadata]] | None = None,
    ) -> None:
        self.artifacts = [
            account.info.artifacts,
            account.info.protocol.artifacts,
            account.avilla.global_artifacts,
        ]
        # 这里是为了能在 Context 层级进行修改

        self.account = account

        self.client = ContextClientSelector.from_selector(self, client)
        self.endpoint = ContextEndpointSelector.from_selector(self, endpoint)
        self.scene = ContextSceneSelector.from_selector(self, scene)
        self.self = ContextSelfSelector.from_selector(self, selft)
        self.mediums = [ContextMedium(ContextSelector.from_selector(self, medium)) for medium in mediums or []]

        self.cache = {"meta": prelude_metadatas or {}}
        self.staff = Staff(self.get_staff_artifacts(), self.get_staff_components())

    @property
    def protocol(self):
        return self.account.info.protocol

    @property
    def avilla(self):
        return self.protocol.avilla

    @property
    def land(self):
        return self.account.info.platform[Land]

    @property
    def request(self) -> ContextRequestSelector:
        return self.endpoint.expects_request()

    @property
    def is_resource(self) -> bool:
        return any(isinstance(i, Resource) for i in self.cache["meta"].get(self.endpoint, {}).values())

    def _collect_metadatas(self, target: Selector | Selectable, *metadatas: Metadata):
        scope = self.cache["meta"].setdefault(target.to_selector(), {})
        scope.update({type(i): i for i in metadatas})

    @classproperty
    @classmethod
    def current(cls) -> Context:
        return cx_context.get()

    def get_staff_components(self):
        return {
            "context": self,
            "protocol": self.protocol,
            "account": self.account,
            "avilla": self.avilla,
        }

    def get_staff_artifacts(self):
        return self.artifacts

    def query(self, pattern: str, **predicators: FollowsPredicater):
        return self.staff.query_entities(pattern, **predicators)

    async def fetch(self, resource: Resource[_T]) -> _T:
        return await self.staff.fetch_resource(resource)

    async def pull(
        self,
        route: type[_MetadataT] | MetadataRoute[Unpack[tuple[Any, ...]], _MetadataT],
        target: Selector | Selectable,
        *,
        flush: bool = False,
    ) -> _MetadataT:
        if isinstance(target, Selectable):
            target = target.to_selector()

        cached = self.cache["meta"].get(target)
        if cached is not None and route in cached:
            meta = cached[route]
            if flush:
                del cached[route]
            if not route.has_params():
                return cast("_MetadataT", meta)

        return await self.staff.pull_metadata(target, route)

    @overload
    def __getitem__(self, closure: Selector) -> ContextSelector:
        ...

    @overload
    def __getitem__(self, closure: Fn[P, R]) -> Callable[P, R]:
        ...

    def __getitem__(self, closure: Selector | Fn[P, Any]):
        if isinstance(closure, Selector):
            return ContextSelector(self, closure.pattern)

        def run(*args: P.args, **kwargs: P.kwargs):
            return self.staff.call_fn(closure, *args, **kwargs)

        return run

    def __staff_generic__(self, element_type: dict, event_type: dict):
        ...



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar, Union

from .endpoint import Endpoint

if TYPE_CHECKING:
    from .perform import BasePerform


T = TypeVar("T")


class Access(Endpoint[T], dynamic=False):
    def evaluate(self, instance: BasePerform) -> T:
        return instance.staff.components[self.name]

    @staticmethod
    def optional():
        return OptionalAccess()


class OptionalAccess(Endpoint[Union[T, None]], dynamic=False):
    def evaluate(self, instance: BasePerform) -> T | None:
        return instance.staff.components.get(self.name)



from .capability import MessageEdit as MessageEdit
from .capability import MessageRevoke as MessageRevoke
from .capability import MessageSend as MessageSend
from .event import MessageEdited as MessageEdited
from .event import MessageReceived as MessageReceived
from .event import MessageRevoked as MessageRevoked
from .event import MessageSent as MessageSent



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, TypedDict

from typing_extensions import ParamSpec, TypeVar

if TYPE_CHECKING:
    from .fn import Fn
    from .overload import FnOverload


P = ParamSpec("P")
R = TypeVar("R", covariant=True)
R1 = TypeVar("R1", covariant=True)

P1 = ParamSpec("P1")
P2 = ParamSpec("P2")
R2 = TypeVar("R2", covariant=True)


class _MergeScopesInfo(TypedDict):
    overload_item: FnOverload
    scopes: list[dict[Any, Any]]


class FnRecord(TypedDict):
    overload_enabled: bool
    overload_scopes: dict[str, dict[Any, Any]]
    record_tuple: Callable | None


@dataclass(eq=True, frozen=True)
class FnImplement:
    fn: Fn

    def merge(self, *records: FnRecord):
        scopes_info: dict[str, _MergeScopesInfo] = {}
        scopes = {}

        for record in records:
            for k, v in record["overload_scopes"].items():
                overload_item = self.fn.overload_map[k]
                info = scopes_info.setdefault(k, {"overload_item": overload_item, "scopes": []})
                info["scopes"].append(v)

        for identity, info in scopes_info.items():
            merged_scope = info["overload_item"].merge_scopes(*info["scopes"])
            scopes[identity] = merged_scope

        return {
            "overload_enabled": self.fn.has_overload_capability,
            "overload_scopes": scopes,
            "record_tuple": records[-1]["record_tuple"],
        }



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Generic,
    TypeVar,
    overload,
)

from typing_extensions import Self

if TYPE_CHECKING:
    from graia.ryanvk.perform import BasePerform


T = TypeVar("T")


class Endpoint(Generic[T]):
    __dynamic__: ClassVar[bool] = False

    name: str

    def __init__(self):
        ...

    def __set_name__(self, owner: type[BasePerform], name: str):
        if self.__dynamic__ and owner.__static__:
            raise TypeError(f"{self.__class__.__name__} conflicts with static perform {owner.__name__}")

        self.name = name

    @overload
    def __get__(self, instance: None, owner: type) -> Self:
        ...

    @overload
    def __get__(self, instance: BasePerform, owner: type) -> T:
        ...

    def __get__(self, instance: BasePerform | None, owner: type):
        if instance is None:
            return self

        return self.evaluate(instance)

    def __init_subclass__(cls, *, dynamic: bool) -> None:
        return super().__init_subclass__()

    def evaluate(self, instance: BasePerform) -> T:
        ...

    @asynccontextmanager
    async def lifespan(self, instance: BasePerform):
        yield



from __future__ import annotations

from typing import Literal

from avilla.core.ryanvk import Capability, Fn, TargetOverload
from avilla.core.selector import Selector


class SceneCapability(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def leave(self, target: Selector) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def disband(self, target: Selector) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def remove_member(self, target: Selector, reason: str | None = None) -> None:
        ...

    # @Fn.with_overload({
    #    TargetOverload(): ['target']
    # })
    # async def request_join(self, solver: Isolate) -> None:
    #     ...

    # TODO: invite someone to join the scene


class RequestJoinCapability(Capability):
    @Fn
    async def on_question(self, target: Selector, question_id: str, question: str, optional: bool) -> str | None:
        ...

    @Fn
    async def on_reason(self, target: Selector) -> str | None:
        ...

    @Fn
    async def on_term(self, term: tuple[Literal["string", "url"] | str, str]) -> bool:
        ...


class RelationshipTerminate(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def terminate(self, target: Selector) -> None:
        # TODO: use wand
        ...



from .capability import BanCapability as BanCapability
from .capability import MuteAllCapability as MuteAllCapability
from .capability import MuteCapability as MuteCapability
from .capability import PrivilegeCapability as PrivilegeCapability
from .metadata import BanInfo as BanInfo
from .metadata import MuteInfo as MuteInfo
from .metadata import Privilege as Privilege



from __future__ import annotations

import re
from copy import deepcopy
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from itertools import filterfalse
from types import MappingProxyType
from typing import Protocol, runtime_checkable

from typing_extensions import Self, TypeAlias

from avilla.core.platform import Land

EMPTY_MAP = MappingProxyType({})
ESCAPE = {")": "\\)", "(": "\\(", "]": "\\]", "[": "\\[", "}": "\\}", "{": "\\{"}

_follows_pattern = re.compile(r"(?P<name>(\w+?|[*~]))(#(?P<predicate>\w+))?(\((?P<literal>[^#]+?)\))?")
FollowsPredicater: TypeAlias = "Callable[[str], bool]"


@dataclass
class _FollowItem:
    name: str
    literal: str | None = None
    predicate: FollowsPredicater | None = None


def _parse_follows_item(item: str, items: dict[str, _FollowItem], predicates: dict[str, FollowsPredicater]):
    if item.startswith("::"):
        if "land" in items:
            raise ValueError("land already exists")
        item = item[2:]
        items["land"] = _FollowItem("land")
        if not item.strip():
            return
    if "*" in items:
        raise ValueError("wildcard already exists, no more items allowed")
    if not (m := _follows_pattern.fullmatch(item)):
        raise ValueError(f"invalid item: {item}")
    name = m["name"]
    if name == "~" and items:
        raise ValueError("inherit can only be the first item and only one")
    if m["literal"] and m["predicate"]:
        raise ValueError(f"duplicate literal and predicate: {item}")
    if m["literal"] is not None:
        literal = m["literal"].translate(str.maketrans(ESCAPE))
    else:
        literal = None
    predicate = predicates.get(m["predicate"]) if m["predicate"] else None
    items[name] = _FollowItem(name, literal, predicate)


def _parse_follows(pattern: str, **kwargs: FollowsPredicater) -> list[_FollowItem]:
    items = {}
    item = ""
    bracket_stack = []
    for i, char in enumerate(pattern):
        if char == "." and not bracket_stack:
            _parse_follows_item(item, items, kwargs)
            item = ""
            continue
        if char == "(":
            bracket_stack.append(i)
        elif char == ")":
            if not bracket_stack:
                raise ValueError(f"Unclosed parenthesis: {item})")
            bracket_stack.pop()
        item += char
    if bracket_stack:
        raise ValueError(f"Unclosed parenthesis: {pattern[bracket_stack[0]:]}")
    if item:
        _parse_follows_item(item, items, kwargs)
    return list(items.values())


class Selector:
    pattern: Mapping[str, str]

    def __init__(self, pattern: Mapping[str, str] = EMPTY_MAP) -> None:
        self.pattern = MappingProxyType({k: str(v) for k, v in pattern.items()})
        self._hash = hash(("Selector", *self.pattern.items()))

    def modify(self, pattern: Mapping[str, str]) -> Self:
        return self.__class__(pattern=pattern)

    def __getattr__(self, name: str) -> Callable[[str], Self]:
        if name.startswith("__"):
            return super().__getattribute__(name)  # type: ignore

        def wrapper(content: str) -> Self:
            return self.modify({**self.pattern, name: str(content)})

        return wrapper

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, o: object) -> bool:
        return isinstance(o, self.__class__) and o._hash == self._hash

    def __contains__(self, key: str) -> bool:
        return key in self.pattern

    def __getitem__(self, key: str) -> str:
        return self.pattern[key]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}().{'.'.join(f'{k}({v})' for k, v in self.pattern.items())}"

    def __copy__(self):
        return self.modify({**self.pattern})

    def __deepcopy__(self, memo):
        data = {**self.pattern}
        return self.__class__(deepcopy(data, memo))

    @property
    def empty(self) -> bool:
        return not self.pattern

    @property
    def path(self) -> str:
        return ".".join(self.pattern)

    @property
    def path_without_land(self) -> str:
        return ".".join(filterfalse(lambda x: x == "land", self.pattern))

    @property
    def display(self) -> str:
        return ".".join(f"{k}({v})" for k, v in self.pattern.items())

    @property
    def display_without_land(self) -> str:
        return ".".join(f"{k}({v})" for k, v in self.pattern.items() if k != "land")

    @property
    def last_key(self) -> str:
        return next(reversed(self.pattern.keys()))

    @property
    def last_value(self) -> str:
        return next(reversed(self.pattern.values()))

    def items(self):
        return self.pattern.items()

    def appendix(self, key: str, value: str):
        return self.modify({**self.pattern, key: str(value)})

    def land(self, land: Land | str):
        if isinstance(land, Land):
            land = land.name

        return self.modify({"land": land, **{k: v for k, v in self.pattern.items() if k != "land"}})

    def to_selector(self):
        return self

    @classmethod
    def from_follows(cls, pattern: str):
        items = _parse_follows(pattern)
        mapping = {}
        for i in items:
            if i.literal is None:
                raise ValueError("literal expected")
            mapping[i.name] = i.literal
        return cls(mapping)

    from_follows_pattern = from_follows

    def follows(self, pattern: str, **kwargs: FollowsPredicater) -> bool:
        items = _parse_follows(pattern, **kwargs)
        index = 0
        for index, (item, name, value) in enumerate(zip(items, self.pattern.keys(), self.pattern.values())):
            if item.name == "*":
                return True
            if item.name != name:
                return False
            if item.predicate is not None and not item.predicate(value):
                return False
            if item.literal is not None and value != item.literal:
                return False
        return index + 1 == len(self.pattern)

    def into(self, pattern: str, **kwargs: str) -> Self:
        items = _parse_follows(pattern)
        new_patterns = {}
        iterator = iter(self.pattern)
        if items and items[0].name == "~":
            if not all(item.literal or kwargs.get(item.name) for item in items[1:]):
                raise ValueError("expected specific literals in follows pattern")
            return self.modify(
                {
                    **self.pattern,
                    **{item.name: kwargs.get(item.name) or item.literal for item in items[1:]},
                }  # type: ignore
            )
        for item in items:
            if item.name == "*":
                raise TypeError("expected no wildcard")
            current_key = next(iterator)
            if item.name != current_key:
                raise ValueError(f"expected {'.'.join(new_patterns)}.{item.name}, got {current_key}")
            new_patterns[current_key] = kwargs.get(item.name) or item.literal or self.pattern[current_key]
        return self.modify(new_patterns)

    def expects(
        self,
        pattern: str,
        *,
        exception_type: type[Exception] = ValueError,
        **kwargs: FollowsPredicater,
    ) -> Self:
        if not self.follows(pattern, **kwargs):
            raise exception_type(f"Selector {self} does not follow {pattern}")

        return self


@runtime_checkable
class Selectable(Protocol):
    def to_selector(self) -> Selector:
        ...



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from graia.amnesia.message import MessageChain
from graia.broadcast.entities.signatures import Force

from avilla.core.account import BaseAccount
from avilla.core.event import AvillaEvent

if TYPE_CHECKING:
    from graia.broadcast.interfaces.dispatcher import DispatcherInterface

    from avilla.core.message import Message
    from avilla.core.selector import Selector


@dataclass
class MessageReceived(AvillaEvent):
    message: Message

    class Dispatcher(AvillaEvent.Dispatcher):
        @classmethod
        async def catch(cls, interface: DispatcherInterface[MessageReceived]):
            if interface.annotation is interface.event.message.__class__:
                return interface.event.message
            if interface.annotation is MessageChain:
                return interface.event.message.content
            if interface.name == "sender":
                return interface.event.message.sender
            return await AvillaEvent.Dispatcher.catch(interface)


@dataclass
class MessageSent(AvillaEvent):
    message: Message
    account: BaseAccount

    class Dispatcher(AvillaEvent.Dispatcher):
        @classmethod
        async def catch(cls, interface: DispatcherInterface[MessageSent]):
            if interface.annotation is interface.event.message.__class__:
                return interface.event.message
            if interface.annotation is MessageChain:
                return interface.event.message.content
            if issubclass(interface.annotation, BaseAccount):
                return interface.event.account
            return await AvillaEvent.Dispatcher.catch(interface)


@dataclass
class MessageEdited(AvillaEvent):
    message: Message
    operator: Selector
    past: MessageChain
    current: MessageChain

    class Dispatcher(AvillaEvent.Dispatcher):
        @classmethod
        async def catch(cls, interface: DispatcherInterface[MessageEdited]):
            if interface.annotation is Message:
                return interface.event.message
            if interface.annotation is MessageChain:
                return interface.event.current
            if interface.name == "sender":
                return interface.event.message.sender
            if interface.name == "operator":
                return interface.event.operator
            return await AvillaEvent.Dispatcher.catch(interface)


@dataclass
class MessageRevoked(AvillaEvent):
    message: Selector
    operator: Selector
    sender: Selector | None = None

    class Dispatcher(AvillaEvent.Dispatcher):
        @classmethod
        async def catch(cls, interface: DispatcherInterface[MessageRevoked]):
            if interface.name == "message":
                return interface.event.message
            if interface.name == "operator":
                return interface.event.operator
            if interface.name == "sender":
                return Force(interface.event.sender)
            return await AvillaEvent.Dispatcher.catch(interface)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .overload.target import LookupCollection


def _merge_lookup_collection(self: LookupCollection, other: LookupCollection):
    for key, branches in self.items():
        if (other_branches := other.pop(key, None)) is None:
            continue

        for header, branch in branches.items():
            if (other_branch := other_branches.pop(header, None)) is None:
                continue

            _merge_lookup_collection(branch.levels, other_branch.levels)
            branch.bind |= other_branch.bind

        branches |= other_branches

    self |= other



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from typing_extensions import get_origin

from avilla.core.event import AvillaEvent
from avilla.core.resource import Resource
from avilla.core.selector import Selector

if TYPE_CHECKING:
    from graia.broadcast.interfaces.dispatcher import DispatcherInterface


@dataclass
class ResourceEvent(AvillaEvent):
    resource: Resource

    class Dispatcher(AvillaEvent.Dispatcher):
        @staticmethod
        async def catch(interface: DispatcherInterface[ResourceEvent]):
            if issubclass((get_origin(interface.annotation)) or interface.annotation, Resource):
                return interface.event.resource
            return await AvillaEvent.Dispatcher.catch(interface)


class ResourceAvailable(ResourceEvent):
    pass


class ResourceUnavailable(ResourceEvent):
    pass


@dataclass
class FileUploaded(ResourceAvailable):
    uploader: Selector
    scene: Selector


@dataclass
class FileRemoved(ResourceUnavailable):
    uploader: Selector
    scene: Selector



from .event import FileRemoved as FileRemoved
from .event import FileUploaded as FileUploaded
from .event import ResourceAvailable as ResourceAvailable
from .event import ResourceEvent as ResourceEvent
from .event import ResourceUnavailable as ResourceUnavailable



from __future__ import annotations

from avilla.core.resource import LocalFileResource, RawResource, UrlResource
from avilla.core.ryanvk.collector.application import ApplicationCollector

try:
    from aiohttp import ClientSession

    aio = True
except ImportError:
    ClientSession = None
    aio = False

from .capability import CoreCapability


class CoreResourceFetchPerform((m := ApplicationCollector())._):
    @m.entity(CoreCapability.fetch, resource=LocalFileResource)
    async def fetch_localfile(self, resource: LocalFileResource):
        return resource.file.read_bytes()

    @m.entity(CoreCapability.fetch, resource=RawResource)
    async def fetch_raw(self, resource: RawResource):
        return resource.data

    if aio:

        @m.entity(CoreCapability.fetch, resource=UrlResource)
        async def fetch_url(self, resource: UrlResource):
            async with ClientSession() as session:
                async with session.get(resource.url) as resp:
                    return await resp.read()

    else:

        @m.entity(CoreCapability.fetch, resource=UrlResource)
        async def fetch_url(self, resource: UrlResource):
            raise NotImplementedError("aiohttp is not installed")



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Generic, TypeVar

from graia.ryanvk import Access, BasePerform

from .base import AvillaBaseCollector

if TYPE_CHECKING:
    from avilla.core.protocol import BaseProtocol


TProtocol = TypeVar("TProtocol", bound="BaseProtocol")
TProtocol1 = TypeVar("TProtocol1", bound="BaseProtocol")

T = TypeVar("T")
T1 = TypeVar("T1")


class ProtocolBasedPerformTemplate(BasePerform, native=True):
    __collector__: ClassVar[ProtocolCollector]

    protocol: Access[BaseProtocol] = Access()


class ProtocolCollector(AvillaBaseCollector, Generic[TProtocol]):
    post_applying: bool = False

    def __init__(self):
        super().__init__()

    @property
    def _(self):
        upper = super()._

        class LocalPerformTemplate(
            Generic[TProtocol1],
            ProtocolBasedPerformTemplate,
            upper,
            native=True,
        ):
            protocol: TProtocol1

        return LocalPerformTemplate[TProtocol]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.event import Dispatchable
from graia.broadcast.interfaces.dispatcher import DispatcherInterface

if TYPE_CHECKING:
    from avilla.core.application import Avilla


@dataclass
class AvillaLifecycleEvent(Dispatchable):
    """指示有关应用 (Avilla) 的事件."""

    avilla: Avilla

    class Dispatcher(BaseDispatcher):
        @classmethod
        async def catch(cls, interface: DispatcherInterface[AvillaLifecycleEvent]):
            from avilla.core.application import Avilla

            if interface.annotation is Avilla:
                return interface.event.avilla


class ApplicationClosing(AvillaLifecycleEvent):
    """指示 Avilla 正在关闭."""


class ApplicationClosed(AvillaLifecycleEvent):
    """指示 Avilla 已经关闭."""


class ApplicationPreparing(AvillaLifecycleEvent):
    """指示 Avilla 正在准备."""


class ApplicationReady(AvillaLifecycleEvent):
    """指示 Avilla 已准备完毕."""



from __future__ import annotations

from avilla.core.metadata import Route
from avilla.core.resource import Resource
from avilla.core.ryanvk import Capability, Fn, MetadataOverload, TargetOverload
from avilla.core.selector import Selector


class SummaryCapability(Capability):
    @Fn.complex({TargetOverload(): ["target"], MetadataOverload(): ["route"]})
    async def set_name(self, target: Selector, route: Route, name: str) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"], MetadataOverload(): ["route"]})
    async def unset_name(self, target: Selector, route: Route) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"], MetadataOverload(): ["route"]})
    async def set_description(self, target: Selector, route: Route, description: str) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"], MetadataOverload(): ["route"]})
    async def unset_description(self, target: Selector, route: Route) -> None:
        ...


class NickCapability(Capability):
    @Fn.complex({TargetOverload(): ["target"], MetadataOverload(): ["route"]})
    async def set_name(self, target: Selector, route: Route, name: str) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def set_nickname(self, target: Selector, nickname: str) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def unset_nickname(self, target: Selector) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def set_badge(self, target: Selector, badge: str) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def unset_badge(self, target: Selector) -> None:
        ...


class AvatarFetch(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    def get_avatar(self, target: Selector) -> Resource[bytes]:
        ...



class InaccessibleInterface(Exception):
    pass


class NetworkError(Exception):
    pass


class HttpRequestError(NetworkError):
    status: int
    reason: str

    def __init__(self, status: int, reason: str) -> None:
        self.status = status
        self.reason = reason

    def __repr__(self) -> str:
        return f"<HttpRequestException status={self.status} reason={self.reason}>"


class ParserException(Exception):  # 解析器错误..我希望你永远不会用到这个
    pass


class ActionFailed(Exception):
    pass


class InvalidAuthentication(Exception):
    pass


class UnsupportedOperation(ActionFailed):
    pass


class InvalidOperation(ActionFailed):
    pass


class AccountMuted(Exception):
    pass


class AccountDeleted(Exception):
    pass


class TooLongMessage(Exception):
    pass


class UnknownTarget(Exception):
    pass


class ContextError(Exception):
    pass


class RemoteError(Exception):
    pass


class UnknownError(Exception):
    """其他错误"""


class DeprecatedError(Exception):
    """该接口已弃用."""


def permission_error_message(operator: str, current: str, required: list[str]):
    return (
        f"missing permission: {operator} required "
        + " or ".join(f'"{i}"' for i in required)
        + f" but current is {current}"
    )



from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, TypeVar

_T = TypeVar("_T")
_R_co = TypeVar("_R_co", covariant=True)


def identity(obj: Any) -> str:
    return obj.__name__ if isinstance(obj, type) else obj.__class__.__name__


class classproperty(Generic[_T, _R_co]):
    fget: tuple[classmethod[_T, [], _R_co]]

    def __init__(self, fget: Callable[[Any], _R_co] | classmethod[_T, [], _R_co]) -> None:
        if not isinstance(fget, classmethod):
            fget = classmethod(fget)

        self.fget = (fget,)

    def __get__(self, __obj: _T, __type: type[_T] | None = None, /) -> _R_co:
        return self.fget[0].__get__(__obj, __type)()



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from launart import Launart, Service
from loguru import logger

from avilla.core.utilles.message_cache import MessageCacheDeque
from avilla.standard.core.application import (
    ApplicationClosed,
    ApplicationClosing,
    ApplicationPreparing,
    ApplicationReady,
)

from .graia import AVILLA_ASCII_LOGO, AVILLA_ASCII_RAW_LOGO, log_telemetry

if TYPE_CHECKING:
    from .application import Avilla
    from .selector import Selector


class AvillaService(Service):
    id = "avilla.service"
    supported_interface_types = set()

    avilla: Avilla
    enabled_cache_message: bool
    message_cache: defaultdict[Selector, MessageCacheDeque]

    def __init__(self, avilla: Avilla, cache_size: int):
        self.avilla = avilla
        if cache_size > 0:
            self.enabled_cache_message = True
            self.message_cache = defaultdict(lambda: MessageCacheDeque(cache_size))
        super().__init__()

    @property
    def required(self) -> set[str]:
        return set()

    @property
    def stages(self):
        return {"preparing", "blocking", "cleanup"}

    def get_interface(self, interface_type):
        ...

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            await self.avilla.broadcast.postEvent(ApplicationPreparing(self.avilla))

            logger.info(AVILLA_ASCII_RAW_LOGO, alt=AVILLA_ASCII_LOGO)
            log_telemetry()

            for protocol in self.avilla.protocols:
                logger.info(
                    f"Using platform: {protocol.__class__}",
                    # alt=f"[magenta]Using platform: [/][dark_orange]{protocol.__class__.platform}[/]",
                )

        await self.avilla.broadcast.postEvent(ApplicationReady(self.avilla))

        async with self.stage("blocking"):
            ...  # TODO: 先放着.

        async with self.stage("cleanup"):
            await self.avilla.broadcast.postEvent(ApplicationClosing(self.avilla))

        await self.avilla.broadcast.postEvent(ApplicationClosed(self.avilla))



from __future__ import annotations

from copy import deepcopy, copy
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Awaitable, Callable, TypeVar, overload

from typing_extensions import Concatenate, ParamSpec, Self, Unpack

from avilla.core.metadata import Metadata, MetadataRoute
from avilla.core.ryanvk import Fn
from avilla.core.selector import EMPTY_MAP, Selector
from avilla.standard.core.privilege import Privilege
from avilla.standard.core.profile import Nick, Summary, Avatar

if TYPE_CHECKING:
    from . import Context

R = TypeVar("R", covariant=True)
P = ParamSpec("P")
_MetadataT = TypeVar("_MetadataT", bound=Metadata)


class ContextSelector(Selector):
    context: Context

    def __init__(self, cx: Context, pattern: Mapping[str, str] = EMPTY_MAP) -> None:
        super().__init__(pattern)
        self.context = cx

    def modify(self, pattern: Mapping[str, str]):
        return self.__class__(self.context, pattern)

    def __deepcopy__(self, memo):
        data = {**self.pattern}
        return self.__class__(copy(self.context), deepcopy(data, memo))

    @classmethod
    def from_selector(cls, cx: Context, selector: Selector) -> Self:
        return cls(cx, selector.pattern)

    @overload
    def __getitem__(self, item: str) -> str:
        ...

    @overload
    def __getitem__(self, item: Fn[Concatenate[Selector, P], R]) -> Callable[P, R]:
        ...

    def __getitem__(
        self,
        item: str | Fn[Concatenate[Selector, P], R],
    ) -> str | Callable[P, R]:
        if isinstance(item, str):
            return super().__getitem__(item)

        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            return self.context.staff.call_fn(item, self, *args, **kwargs)

        return wrapper

    def pull(
        self, metadata: type[_MetadataT] | MetadataRoute[Unpack[tuple[Any, ...]], _MetadataT]
    ) -> Awaitable[_MetadataT]:
        return self.context.pull(metadata, self)

    def nick(self):
        return self.pull(Nick)

    def summary(self):
        return self.pull(Summary)

    def avatar(self):
        return self.pull(Avatar)

    def privilege(self):
        return self.pull(Privilege)



from __future__ import annotations

from collections import ChainMap
from itertools import chain
from typing import Any

from graia.ryanvk.typing import SupportsMerge

GLOBAL_GALLERY = {}  # layout: {namespace: {identify: {...}}}, cover-mode.


def ref(namespace: str, identify: str | None = None) -> dict[Any, Any]:
    ns = GLOBAL_GALLERY.setdefault(namespace, {})
    scope = ns.setdefault(identify or "_", {})
    return scope


def merge(*artifacts: dict[Any, Any]):
    chainmap = ChainMap(*artifacts)
    total_signatures = list(dict.fromkeys(chain(*[i.keys() for i in artifacts])))
    result = {}

    for sign in total_signatures:
        if isinstance(sign, SupportsMerge):
            records = [i[sign] for i in artifacts if sign in i]
            result[sign] = sign.merge(*records)
        else:
            result[sign] = chainmap[sign]

    return result



from ._runtime import GLOBAL_GALLERY as GLOBAL_GALLERY
from ._runtime import merge as merge
from ._runtime import ref as ref
from .access import Access as Access
from .access import OptionalAccess as OptionalAccess
from .behavior import OverloadBehavior as OverloadBehavior
from .capability import Capability as Capability
from .collector import BaseCollector as BaseCollector
from .endpoint import Endpoint as Endpoint
from .fn import Fn as Fn
from .gateway import GLOBAL_GATEWAY as GLOBAL_GATEWAY
from .gateway import Gateway as Gateway
from .overload import FnOverload as FnOverload
from .overload import PredicateOverload as PredicateOverload
from .overload import SimpleOverload as SimpleOverload
from .overload import TypeOverload as TypeOverload
from .override import OverridePerformEntity as OverridePerformEntity
from .perform import BasePerform as BasePerform
from .staff import Staff as Staff
from .typing import SupportsCollect as SupportsCollect
from .typing import SupportsMerge as SupportsMerge



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from inspect import isclass
from typing import TYPE_CHECKING, Any, Literal

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.event import Dispatchable
from graia.broadcast.entities.signatures import Force
from typing_extensions import Unpack

from avilla.core.account import BaseAccount
from avilla.core.metadata import FieldReference, Metadata, MetadataRoute
from avilla.core.selector import Selector

from ._runtime import cx_context

if TYPE_CHECKING:
    from graia.broadcast.interfaces.dispatcher import DispatcherInterface

    from .context import Context


@dataclass
class AvillaEvent(Dispatchable):
    context: Context
    time: datetime = field(init=False, default_factory=datetime.now)

    class Dispatcher(BaseDispatcher):
        @staticmethod
        async def beforeExecution(interface: DispatcherInterface[AvillaEvent]):
            if interface.depth < 1:
                interface.local_storage["avilla_context"] = interface.event.context
                interface.local_storage["_context_token"] = cx_context.set(interface.event.context)
            await interface.event.context.staff.exit_stack.__aenter__()

        @staticmethod
        async def catch(interface: DispatcherInterface[AvillaEvent]):
            from .context import Context

            if interface.annotation is Context:
                return interface.event.context

            if isclass(interface.annotation) and issubclass(interface.annotation, BaseAccount):
                return interface.event.context.account

            if interface.name == "client":
                return interface.event.context.client

            if interface.name == "scene":
                return interface.event.context.scene

            if interface.name == "endpoint":
                return interface.event.context.endpoint

        @staticmethod
        async def afterExecution(interface: DispatcherInterface[AvillaEvent], exc, tb):
            if interface.depth < 1:
                cx_context.reset(interface.local_storage["_context_token"])
            await interface.event.context.staff.exit_stack.__aexit__(type(exc), exc, tb)


@dataclass
class RelationshipEvent(AvillaEvent):
    ...


@dataclass
class RelationshipCreated(RelationshipEvent):
    ...


@dataclass
class DirectSessionCreated(RelationshipCreated):
    ...


@dataclass
class SceneCreated(RelationshipCreated):
    ...


@dataclass
class MemberCreated(RelationshipEvent):
    ...


@dataclass
class RelationshipDestroyed(RelationshipEvent):
    active: bool
    indirect: bool = False


@dataclass
class DirectSessionDestroyed(RelationshipDestroyed):
    ...


@dataclass
class SceneDestroyed(RelationshipDestroyed):
    ...


@dataclass
class MemberDestroyed(RelationshipDestroyed):
    ...


@dataclass
class ModifyDetail:
    type: Literal["set", "clear", "update"]
    current: Any = None
    previous: Any = None


@dataclass
class MetadataModified(AvillaEvent):
    endpoint: Selector
    route: type[Metadata] | MetadataRoute[Unpack[tuple[Any, ...]], Metadata]
    details: dict[FieldReference, ModifyDetail]
    operator: Selector | None = None
    scene: Selector | None = None

    class Dispatcher(AvillaEvent.Dispatcher):
        @staticmethod
        async def catch(interface: DispatcherInterface["MetadataModified"]):
            if interface.name == "route":
                return interface.event.route
            if interface.name == "details":
                return interface.event.details
            if interface.name == "operator":
                return Force(interface.event.operator)
            return await AvillaEvent.Dispatcher.catch(interface)



from __future__ import annotations

from dataclasses import dataclass

from graia.broadcast.interfaces.dispatcher import DispatcherInterface

from avilla.core.event import AvillaEvent
from avilla.core.selector import Selector


@dataclass
class ActivityEvent(AvillaEvent):
    id: str  # eg. "button_clicked"
    scene: Selector
    activity: Selector  # eg. [...].button("#1")

    def to_selector(self):  # eg. [...].activity("button_clicked")
        return self.scene.activity(self.id)

    class Dispatcher(AvillaEvent.Dispatcher):
        @classmethod
        async def catch(cls, interface: DispatcherInterface[ActivityEvent]):
            if interface.name == "activity":
                return interface.event.to_selector()
            return await AvillaEvent.Dispatcher.catch(interface)


@dataclass
class ActivityAvailable(ActivityEvent):
    pass


@dataclass
class ActivityUnavailable(ActivityEvent):
    pass


@dataclass
class ActivityTrigged(ActivityEvent):
    trigger: Selector  # who trigged the activity

    class Dispatcher(ActivityEvent.Dispatcher):
        @classmethod
        async def catch(cls, interface: DispatcherInterface[ActivityTrigged]):
            if interface.name == "trigger":
                return interface.event.trigger
            return await ActivityEvent.Dispatcher.catch(interface)



from __future__ import annotations

from dataclasses import dataclass

from avilla.core.metadata import Metadata


@dataclass
class Summary(Metadata):
    name: str
    description: str | None


@dataclass
class Nick(Metadata):
    name: str
    nickname: str
    badge: str | None


@dataclass
class Avatar(Metadata):
    url: str



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta

from avilla.core.metadata import Metadata


@dataclass
class MuteInfo(Metadata):
    muted: bool
    duration: timedelta
    time: datetime | None = None


@dataclass
class BanInfo(Metadata):
    banned: bool
    duration: timedelta
    time: datetime | None = None


@dataclass
class Privilege(Metadata):
    available: bool
    effective: bool

    duration: timedelta | None = None



from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING
from weakref import WeakValueDictionary

if TYPE_CHECKING:
    from avilla.core.message import Message
    from avilla.core.selector import Selector


class MessageCacheDeque:
    messages: deque[Message]
    _endpoint: WeakValueDictionary[Selector, Message]

    def __init__(self, cache_size: int) -> None:
        self.messages = deque(maxlen=cache_size)
        self._endpoint = WeakValueDictionary()

    def push(self, message: Message):
        self.messages.append(message)
        self._endpoint[message.to_selector()] = message

    def get(self, message_selector: Selector):
        return self._endpoint.get(message_selector)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from statv import Stats, Statv

from avilla.core.ryanvk.staff import Staff
from avilla.core.selector import Selector

if TYPE_CHECKING:
    from avilla.core.application import Avilla
    from avilla.core.context import Context
    from avilla.core.platform import Platform
    from avilla.core.protocol import BaseProtocol


@dataclass
class AccountInfo:
    route: Selector
    account: BaseAccount
    protocol: BaseProtocol
    platform: Platform
    artifacts: dict[Any, Any] = field(default_factory=dict)
    enabled_message_cache: bool = False


@dataclass
class BaseAccount:
    route: Selector
    avilla: Avilla

    @property
    def info(self) -> AccountInfo:
        return self.avilla.accounts[self.route]

    @property
    def staff(self):
        return Staff(self.get_staff_artifacts(), self.get_staff_components())

    @property
    def available(self) -> bool:
        return True

    def get_context(self, target: Selector, *, via: Selector | None = None) -> Context:
        return self.staff.get_context(target, via=via)

    def get_self_context(self):
        from avilla.core.context import Context

        return Context(
            self,
            self.route,
            self.route,
            self.route.into("::"),
            self.route,
        )

    def get_staff_components(self):
        return {"account": self, "protocol": self.info.protocol, "avilla": self.avilla}

    def get_staff_artifacts(self):
        return [
            self.info.artifacts,
            self.info.protocol.artifacts,
            self.avilla.global_artifacts,
        ]

    def __staff_generic__(self, element_type: dict, event_type: dict):
        ...


class AccountStatus(Statv):
    enabled = Stats[bool]("enabled", default=False)

    @property
    def available(self) -> bool:
        return self.enabled



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING

from avilla.standard.core.request.capability import RequestCapability

from ._runtime import cx_context
from .metadata import Metadata
from .platform import Land

if TYPE_CHECKING:
    from .account import BaseAccount
    from .selector import Selector


@dataclass
class Request(Metadata):
    id: str
    land: Land
    scene: Selector
    sender: Selector
    account: BaseAccount
    time: datetime

    request_type: str | None = None
    operator: Selector | None = None
    response: str | None = None

    @property
    def solved(self) -> bool:
        return self.response is not None

    @property
    def pending(self) -> bool:
        return self.response is None

    @property
    def accepted(self) -> bool:
        return self.response == "accept"

    @property
    def rejected(self) -> bool:
        return self.response == "reject"

    @property
    def cannelled(self) -> bool:
        return self.response == "cancel"

    @property
    def ignored(self) -> bool:
        return self.response == "ignore"

    def to_selector(self) -> Selector:
        request_id = self.id
        if self.request_type is not None:
            request_id = f"{self.request_type}:{request_id}"
        return self.scene.request(request_id)

    async def accept(self):
        return await cx_context.get()[RequestCapability.accept](self.to_selector())

    async def reject(self, reason: str | None = None, forever: bool = False):
        return await cx_context.get()[RequestCapability.reject](self.to_selector(), reason, forever)

    async def cancel(self):
        return await cx_context.get()[RequestCapability.cancel](self.to_selector())

    async def ignore(self):
        return await cx_context.get()[RequestCapability.ignore](self.to_selector())



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from graia.amnesia.message import Element
from graia.amnesia.message.element import Text as Text  # noqa
from graia.amnesia.message.element import Unknown as Unknown  # noqa

from avilla.core.resource import LocalFileResource, Resource
from avilla.core.selector import Selector


class Notice(Element):
    """该消息元素用于承载消息中用于提醒/呼唤特定用户的部分."""

    target: Selector
    display: str | None

    def __init__(self, target: Selector, display: str | None = None) -> None:
        """实例化一个 Notice 消息元素，用于承载消息中用于提醒/呼唤特定用户的部分。

        Args:
            target (str): 需要提醒/呼唤的特定用户的 ID.
            display (str, optional): 需要提醒/呼唤的特定用户的显示名称. Defaults to None.
        """

        self.target = target
        self.display = display

    def __str__(self) -> str:
        return "[$Notice]"

    def __repr__(self) -> str:
        return f"[$Notice:target={self.target}]"


class NoticeAll(Element):
    """该消息元素用于群组中的管理员提醒群组中的所有成员"""

    def __init__(self) -> None:
        super().__init__()

    def __str__(self) -> str:
        return "[$NoticeAll]"

    def __eq__(self, other):
        return isinstance(other, NotImplementedError)


class Picture(Element):
    resource: Resource[bytes]

    def __init__(self, resource: Resource[bytes] | Path | str):
        if isinstance(resource, Path):
            resource = LocalFileResource(resource)
        elif isinstance(resource, str):
            resource = LocalFileResource(Path(resource))
        self.resource = resource

    def __str__(self) -> str:
        return "[$Picture]"

    def __repr__(self):
        return f"[$Picture:resource={self.resource.to_selector()}]"

    def __eq__(self, other):
        return isinstance(other, Picture) and self.resource.to_selector() == other.resource.to_selector()


class Audio(Element):
    resource: Resource[bytes]
    duration: int

    def __init__(self, resource: Resource[bytes] | Path | str, duration: int = -1):
        if isinstance(resource, Path):
            resource = LocalFileResource(resource)
        elif isinstance(resource, str):
            resource = LocalFileResource(Path(resource))
        self.resource = resource
        self.duration = duration

    def __str__(self) -> str:
        return "[$Audio]"

    def __repr__(self):
        return f"[$Audio:resource={self.resource.to_selector()}]"

    def __eq__(self, other):
        return isinstance(other, Audio) and self.resource.to_selector() == other.resource.to_selector()


class Video(Element):
    resource: Resource[bytes]

    def __init__(self, resource: Resource[bytes] | Path | str):
        if isinstance(resource, Path):
            resource = LocalFileResource(resource)
        elif isinstance(resource, str):
            resource = LocalFileResource(Path(resource))
        self.resource = resource

    def __str__(self) -> str:
        return "[$Video]"

    def __repr__(self):
        return f"[$Video:resource={self.resource.to_selector()}]"

    def __eq__(self, other):
        return isinstance(other, Video) and self.resource.to_selector() == other.resource.to_selector()


class File(Element):
    resource: Resource[bytes]

    def __init__(self, resource: Resource[bytes] | Path | str):
        if isinstance(resource, Path):
            resource = LocalFileResource(resource)
        elif isinstance(resource, str):
            resource = LocalFileResource(Path(resource))
        self.resource = resource

    def __str__(self) -> str:
        return "[$File]"

    def __repr__(self) -> str:
        return f"[$File:resource={self.resource.to_selector()}]"

    def __eq__(self, other):
        return isinstance(other, File) and self.resource.to_selector() == other.resource.to_selector()


@dataclass
class Reference(Element):
    message: Selector
    slice: tuple[int, int] | None = None

    def __str__(self):
        return f"[$Reference:id={self.message}]"

    def __eq__(self, other):
        return isinstance(other, Reference) and self.message == other.message


@dataclass
class Face(Element):
    id: str
    name: str | None = None

    def __str__(self) -> str:
        return f"[Face:id={self.id};name={self.name}]"



from __future__ import annotations

from inspect import isclass
from typing import TYPE_CHECKING

from graia.broadcast.entities.dispatcher import BaseDispatcher

from avilla.core._runtime import cx_protocol
from avilla.core.account import BaseAccount
from avilla.core.context import Context
from avilla.core.event import AvillaEvent
from avilla.core.protocol import BaseProtocol

if TYPE_CHECKING:
    from graia.broadcast.interfaces.dispatcher import DispatcherInterface

    from avilla.core.application import Avilla


class AvillaBuiltinDispatcher(BaseDispatcher):
    avilla: Avilla

    def __init__(self, avilla: Avilla) -> None:
        self.avilla = avilla

    async def catch(self, interface: DispatcherInterface[AvillaEvent]):
        from avilla.core.application import Avilla

        if interface.annotation is Avilla:
            return self.avilla
        if interface.annotation in self.avilla._protocol_map:
            return self.avilla._protocol_map[interface.annotation]
        if (
            isclass(interface.annotation)
            and issubclass(interface.annotation, BaseProtocol)
            and isinstance(cx_protocol.get(None), interface.annotation)
        ):
            return cx_protocol.get(None)
        if (
            isinstance(interface.event, AvillaEvent)
            and isclass(interface.annotation)
            and issubclass(interface.annotation, BaseAccount)
        ):
            cx: Context = interface.local_storage["avilla_context"]
            return cx.account



"""
参考路径如下：

## Cache path:
    macOS: ~/Library/Caches/<AppName>

    Unix: ~/.cache/<AppName> (XDG default)

    Windows: C:\\Users\\<username>\\AppData\\Local\\<AppName>\\Cache

## Data path:
    macOS: ~/Library/Application Support/<AppName>

    Unix: ~/.local/share/<AppName> or in $XDG_DATA_HOME, if defined

    Win XP (not roaming): C:\\Documents and Settings\\<username>\\Application Data\\<AppName>

    Win 7 (not roaming): C:\\Users\\<username>\\AppData\\Local\\<AppName>

## Config path:
    macOS: same as user_data_dir

    Unix: ~/.config/<AppName>

    Win XP (roaming): C:\\Documents and Settings\\<username>\\Local Settings\\Application Data\\<AppName>

    Win 7 (roaming): C:\\Users\\<username>\\AppData\\Roaming\\<AppName>

"""

import os
import sys
from pathlib import Path
from typing import Callable, Literal, Optional

from typing_extensions import ParamSpec

WINDOWS = sys.platform.startswith("win") or (sys.platform == "cli" and os.name == "nt")


def user_cache_dir(appname: str) -> Path:
    r"""
    Return full path to the user-specific cache dir for this application.
        "appname" is the name of application.
    Typical user cache directories are:
        macOS:      ~/Library/Caches/<AppName>
        Unix:       ~/.cache/<AppName> (XDG default)
        Windows:    C:\\Users\\<username>\\AppData\\Local\\<AppName>\\Cache
    On Windows the only suggestion in the MSDN docs is that local settings go
    in the `CSIDL_LOCAL_APPDATA` directory. This is identical to the
    non-roaming app data dir (the default returned by `user_data_dir`). Apps
    typically put cache data somewhere *under* the given dir here. Some
    examples:
        ...\\Mozilla\\Firefox\\Profiles\\<ProfileName>\\Cache
        ...\\Acme\\SuperApp\\Cache\\1.0
    OPINION: This function appends "Cache" to the `CSIDL_LOCAL_APPDATA` value.
    """
    if WINDOWS:
        return _get_win_folder("CSIDL_LOCAL_APPDATA") / appname / "Cache"
    elif sys.platform == "darwin":
        return Path("~/Library/Caches").expanduser() / appname
    else:
        return Path(os.getenv("XDG_CACHE_HOME", "~/.cache")).expanduser() / appname


def user_data_dir(appname: str, roaming: bool = False) -> Path:
    r"""
    Return full path to the user-specific data dir for this application.
        "appname" is the name of application.
            If None, just the system directory is returned.
        "roaming" (boolean, default False) can be set True to use the Windows
            roaming appdata directory. That means that for users on a Windows
            network setup for roaming profiles, this user data will be
            sync'd on login. See
            <http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx>
            for a discussion of issues.
    Typical user data directories are:
        macOS:                  ~/Library/Application Support/<AppName>
        Unix:                   ~/.local/share/<AppName>    # or in
                                $XDG_DATA_HOME, if defined
        Win XP (not roaming):   C:\\Documents and Settings\\<username>\\ ...
                                ...Application Data\\<AppName>
        Win XP (roaming):       C:\\Documents and Settings\\<username>\\Local ...
                                ...Settings\\Application Data\\<AppName>
        Win 7  (not roaming):   C:\\Users\\<username>\\AppData\\Local\\<AppName>
        Win 7  (roaming):       C:\\Users\\<username>\\AppData\\Roaming\\<AppName>
    For Unix, we follow the XDG spec and support $XDG_DATA_HOME.
    That means, by default "~/.local/share/<AppName>".
    """
    if WINDOWS:
        const = "CSIDL_APPDATA" if roaming else "CSIDL_LOCAL_APPDATA"
        return Path(_get_win_folder(const)) / appname  # type: ignore
    elif sys.platform == "darwin":
        return Path("~/Library/Application Support/").expanduser() / appname
    else:
        return Path(os.getenv("XDG_DATA_HOME", "~/.local/share")).expanduser() / appname


def user_config_dir(appname: str, roaming: bool = True) -> Path:
    """Return full path to the user-specific config dir for this application.
        "appname" is the name of application.
            If None, just the system directory is returned.
        "roaming" (boolean, default True) can be set False to not use the
            Windows roaming appdata directory. That means that for users on a
            Windows network setup for roaming profiles, this user data will be
            sync'd on login. See
            <http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx>
            for a discussion of issues.
    Typical user data directories are:
        macOS:                  same as user_data_dir
        Unix:                   ~/.config/<AppName>
        Win *:                  same as user_data_dir
    For Unix, we follow the XDG spec and support $XDG_CONFIG_HOME.
    That means, by default "~/.config/<AppName>".
    """
    if WINDOWS:
        return user_data_dir(appname, roaming=roaming)
    elif sys.platform == "darwin":
        return user_data_dir(appname)
    else:
        return Path(os.getenv("XDG_CONFIG_HOME", "~/.config")).expanduser() / appname


# -- Windows support functions --
def _get_win_folder_from_registry(
    csidl_name: Literal["CSIDL_APPDATA", "CSIDL_COMMON_APPDATA", "CSIDL_LOCAL_APPDATA"]
) -> Path:
    """
    This is a fallback technique at best. I'm not sure if using the
    registry for this guarantees us the correct answer for all CSIDL_*
    names.
    """
    import winreg

    shell_folder_name = {
        "CSIDL_APPDATA": "AppData",
        "CSIDL_COMMON_APPDATA": "Common AppData",
        "CSIDL_LOCAL_APPDATA": "Local AppData",
    }[csidl_name]

    key = winreg.OpenKey(
        winreg.HKEY_CURRENT_USER,
        r"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders",
    )
    directory, _type = winreg.QueryValueEx(key, shell_folder_name)
    return Path(directory)


def _get_win_folder_with_ctypes(
    csidl_name: Literal["CSIDL_APPDATA", "CSIDL_COMMON_APPDATA", "CSIDL_LOCAL_APPDATA"]
) -> Path:
    import ctypes

    csidl_const = {
        "CSIDL_APPDATA": 26,
        "CSIDL_COMMON_APPDATA": 35,
        "CSIDL_LOCAL_APPDATA": 28,
    }[csidl_name]

    buf = ctypes.create_unicode_buffer(1024)
    ctypes.windll.shell32.SHGetFolderPathW(None, csidl_const, None, 0, buf)

    # Downgrade to short path name if have highbit chars. See
    # <http://bugs.activestate.com/show_bug.cgi?id=85099>.
    has_high_char = any(ord(c) > 255 for c in buf)
    if has_high_char:
        buf2 = ctypes.create_unicode_buffer(1024)
        if ctypes.windll.kernel32.GetShortPathNameW(buf.value, buf2, 1024):
            buf = buf2

    return Path(buf.value)


if WINDOWS:
    try:
        _get_win_folder = _get_win_folder_with_ctypes
    except ImportError:
        _get_win_folder = _get_win_folder_from_registry


P = ParamSpec("P")

APP_NAME = "avilla"
BASE_CACHE_DIR = user_cache_dir(APP_NAME).resolve()
BASE_CONFIG_DIR = user_config_dir(APP_NAME).resolve()
BASE_DATA_DIR = user_data_dir(APP_NAME).resolve()


def _ensure_dir(path: Path) -> None:
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    elif not path.is_dir():
        raise RuntimeError(f"{path} is not a directory")


def _auto_create_dir(func: Callable[P, Path]) -> Callable[P, Path]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Path:
        path = func(*args, **kwargs)
        _ensure_dir(path)
        return path

    return wrapper


@_auto_create_dir
def get_cache_dir(plugin_name: Optional[str]) -> Path:
    """
    macOS: ~/Library/Caches/<AppName>

    Unix: ~/.cache/<AppName> (XDG default)

    Windows: C:\\Users\\<username>\\AppData\\Local\\<AppName>\\Cache
    """
    return BASE_CACHE_DIR / plugin_name if plugin_name else BASE_CACHE_DIR


def get_cache_file(plugin_name: Optional[str], filename: str) -> Path:
    """
    macOS: ~/Library/Caches/<AppName>

    Unix: ~/.cache/<AppName> (XDG default)

    Windows: C:\\Users\\<username>\\AppData\\Local\\<AppName>\\Cache
    """
    return get_cache_dir(plugin_name) / filename


@_auto_create_dir
def get_config_dir(plugin_name: Optional[str]) -> Path:
    """
    macOS: same as user_data_dir

    Unix: ~/.config/<AppName>

    Win XP (roaming): C:\\Documents and Settings\\<username>\\Local Settings\\Application Data\\<AppName>

    Win 7 (roaming): C:\\Users\\<username>\\AppData\\Roaming\\<AppName>
    """
    return BASE_CONFIG_DIR / plugin_name if plugin_name else BASE_CONFIG_DIR


def get_config_file(plugin_name: Optional[str], filename: str) -> Path:
    """
    macOS: same as user_data_dir

    Unix: ~/.config/<AppName>

    Win XP (roaming): C:\\Documents and Settings\\<username>\\Local Settings\\Application Data\\<AppName>

    Win 7 (roaming): C:\\Users\\<username>\\AppData\\Roaming\\<AppName>
    """
    return get_config_dir(plugin_name) / filename


@_auto_create_dir
def get_data_dir(plugin_name: Optional[str]) -> Path:
    """
    macOS: ~/Library/Application Support/<AppName>

    Unix: ~/.local/share/<AppName> or in $XDG_DATA_HOME, if defined

    Win XP (not roaming): C:\\Documents and Settings\\<username>\\Application Data\\<AppName>

    Win 7 (not roaming): C:\\Users\\<username>\\AppData\\Local\\<AppName>
    """
    return BASE_DATA_DIR / plugin_name if plugin_name else BASE_DATA_DIR


def get_data_file(plugin_name: Optional[str], filename: str) -> Path:
    """
    macOS: ~/Library/Application Support/<AppName>

    Unix: ~/.local/share/<AppName> or in $XDG_DATA_HOME, if defined

    Win XP (not roaming): C:\\Documents and Settings\\<username>\\Application Data\\<AppName>

    Win 7 (not roaming): C:\\Users\\<username>\\AppData\\Local\\<AppName>
    """
    return get_data_dir(plugin_name) / filename



from __future__ import annotations

from avilla.core.ryanvk import Capability, Fn, TargetOverload
from avilla.core.selector import Selector


class ActivityTrigger(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def trigger(self, target: Selector):
        ...



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Generic, TypeVar

from graia.ryanvk import Access, BasePerform

from .base import AvillaBaseCollector

if TYPE_CHECKING:
    from avilla.core.account import BaseAccount
    from avilla.core.context import Context
    from avilla.core.metadata import Metadata
    from avilla.core.protocol import BaseProtocol


TProtocol = TypeVar("TProtocol", bound="BaseProtocol")
TAccount = TypeVar("TAccount", bound="BaseAccount")

TProtocol1 = TypeVar("TProtocol1", bound="BaseProtocol")
TAccount1 = TypeVar("TAccount1", bound="BaseAccount")

T = TypeVar("T")
T1 = TypeVar("T1")
M = TypeVar("M", bound="Metadata")


class ContextBasedPerformTemplate(BasePerform, native=True):
    __collector__: ClassVar[ContextCollector]

    context: Access[Context] = Access()

    @property
    def protocol(self):
        return self.context.protocol

    @property
    def account(self):
        return self.context.account


class ContextCollector(AvillaBaseCollector, Generic[TProtocol, TAccount]):
    post_applying: bool = False

    def __init__(self):
        super().__init__()

    @property
    def _(self):
        upper = super()._

        class LocalPerformTemplate(
            Generic[TProtocol1, TAccount1],
            ContextBasedPerformTemplate,
            upper,
            native=True,
        ):
            protocol: TProtocol1
            account: TAccount1

        return LocalPerformTemplate[TProtocol, TAccount]



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime

from graia.amnesia.message import MessageChain

from avilla.core.platform import Land
from avilla.core.selector import Selector
from avilla.standard.core.message.capability import MessageRevoke

from ._runtime import cx_context
from .metadata import Metadata


@dataclass
class Message(Metadata):
    id: str
    scene: Selector
    sender: Selector
    content: MessageChain
    time: datetime
    reply: Selector | None = None

    @property
    def land(self):
        return Land(self.scene["land"])

    def to_selector(self) -> Selector:
        return self.scene.message(self.id)

    async def revoke(self):
        await cx_context.get()[MessageRevoke.revoke](self.to_selector())



from .event import ApplicationClosed as ApplicationClosed
from .event import ApplicationClosing as ApplicationClosing
from .event import ApplicationPreparing as ApplicationPreparing
from .event import ApplicationReady as ApplicationReady
from .event import AvillaLifecycleEvent as AvillaLifecycleEvent



from .capability import ActivityTrigger as ActivityTrigger
from .event import ActivityAvailable as ActivityAvailable
from .event import ActivityEvent as ActivityEvent
from .event import ActivityTrigged as ActivityTrigged
from .event import ActivityUnavailable as ActivityUnavailable



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Callable, Generic, overload

from typing_extensions import Concatenate, ParamSpec, Self, TypeVar

from graia.ryanvk.sign import FnImplement

from .behavior import DEFAULT_BEHAVIOR, OverloadBehavior
from .override import OverridePerformEntity
from .perform import BasePerform

if TYPE_CHECKING:
    from .capability import Capability
    from .collector import BaseCollector
    from .overload import FnOverload
    from .staff import Staff
    from .typing import SupportsCollect

T = TypeVar("T")

P = ParamSpec("P")
P1 = ParamSpec("P1")
P2 = ParamSpec("P2")

R = TypeVar("R", covariant=True)
R1 = TypeVar("R1", covariant=True)
R2 = TypeVar("R2", covariant=True)

N = TypeVar("N", bound="BasePerform")

VnCallable = TypeVar("VnCallable", bound=Callable, covariant=True)


class Fn(Generic[P, R]):
    owner: type[BasePerform | Capability]  # TODO: review here
    name: str

    shape: Callable
    shape_signature: inspect.Signature

    behavior: OverloadBehavior

    overload_params: dict[str, FnOverload]
    overload_param_map: dict[FnOverload, list[str]]
    overload_map: dict[str, FnOverload]

    def __init__(
        self: Fn[P, R],
        shape: Callable[Concatenate[Any, P], R],
        behavior: OverloadBehavior = DEFAULT_BEHAVIOR,
        overload_param_map: dict[FnOverload, list[str]] | None = None,
    ):
        self.shape = shape
        self.shape_signature = inspect.Signature(list(inspect.Signature.from_callable(shape).parameters.values())[1:])
        self.behavior = behavior

        self.overload_param_map = overload_param_map or {}
        self.overload_params = {i: k for k, v in self.overload_param_map.items() for i in v}
        self.overload_map = {i.identity: i for i in self.overload_param_map}

    def __set_name__(self, owner: type[BasePerform], name: str):
        self.owner = owner
        self.name = name

    @overload
    def __get__(self, instance: BasePerform, owner: type) -> Callable[P, R]:
        ...

    @overload
    def __get__(self, instance: Any, owner: type) -> Self:
        ...

    def __get__(self, instance: BasePerform | None, owner: type):
        if not isinstance(instance, BasePerform):
            return self

        def wrapper(*args: P.args, **kwargs: P.kwargs):
            return instance.staff.call_fn(self, *args, **kwargs)

        return wrapper

    @classmethod
    def complex(cls, overload_param_map: dict[FnOverload, list[str]], behavior: OverloadBehavior = DEFAULT_BEHAVIOR):
        def wrapper(shape: Callable[Concatenate[Any, P1], R1]) -> Fn[P1, R1]:
            return cls(shape, overload_param_map=overload_param_map, behavior=behavior)  # type: ignore

        return wrapper

    @property
    def has_overload_capability(self) -> bool:
        return bool(self.overload_param_map)

    def collect(
        self,
        collector: BaseCollector,
        **overload_settings: Any,
    ):
        def wrapper(entity: Callable[Concatenate[Any, P], R]) -> Callable[Concatenate[Any, P], R]:
            artifact = collector.artifacts.setdefault(
                FnImplement(self),
                {
                    "overload_enabled": self.has_overload_capability,
                    "overload_scopes": {},
                    "record_tuple": None,
                },
            )
            if self.has_overload_capability:
                for fn_overload, params in self.overload_param_map.items():
                    scope = artifact["overload_scopes"].setdefault(fn_overload.identity, {})
                    fn_overload.collect_entity(
                        collector,
                        scope,
                        entity,
                        fn_overload.get_params_layout(params, overload_settings),
                    )
            else:
                artifact["record_tuple"] = (collector, entity)

            return entity

        return wrapper

    def dyn_perform(self, staff: Staff, cls: type[N]) -> N:
        raise ValueError("common staff can only works with static perform")

    def _get_instance(self, staff: Staff, cls: type[N]) -> N:
        if cls not in staff.instances:
            if cls.__static__:
                instance = staff.instances[cls] = cls(staff)
            else:
                instance = self.dyn_perform(staff, cls)
        else:
            instance = staff.instances[cls]

        return instance

    def execute(
        self,
        staff: Staff,
        collector: BaseCollector,
        entity: Callable[Concatenate[Any, P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        # get instance
        instance = self._get_instance(staff, collector.cls)

        # execute
        return entity(instance, *args, **kwargs)

    def override(
        self: SupportsCollect[P1, Callable[[Callable[Concatenate[Any, P2], R2]], Any]],  # pyright: ignore
        collector: BaseCollector,
        *args: P1.args,
        **kwargs: P1.kwargs,
    ):
        def wrapper(entity: Callable[Concatenate[Any, P2], R2]) -> OverridePerformEntity[P2, R2]:
            self.collect(collector, *args, **kwargs)(entity)
            return OverridePerformEntity(collector, self, entity)  # type: ignore

        return wrapper



from __future__ import annotations

import asyncio
import signal
from typing import TYPE_CHECKING, Any, Callable, Iterable, TypeVar, overload

from creart import it
from graia.amnesia.builtins.memcache import MemcacheService
from graia.broadcast import Broadcast
from launart import Launart
from launart.service import Service
from loguru import logger

from avilla.core._runtime import get_current_avilla
from avilla.core.account import AccountInfo, BaseAccount
from avilla.core.dispatchers import AvillaBuiltinDispatcher
from avilla.core.protocol import BaseProtocol
from avilla.core.ryanvk.staff import Staff
from avilla.core.selector import Selector
from avilla.core.service import AvillaService
from avilla.core.utilles import identity

if TYPE_CHECKING:
    from graia.broadcast import Decorator, Dispatchable, Namespace, T_Dispatcher

    from avilla.core.event import AvillaEvent
    from avilla.standard.core.application import AvillaLifecycleEvent

    from .resource import Resource

T = TypeVar("T")
TE = TypeVar("TE", bound="AvillaEvent")
TP = TypeVar("TP", bound=BaseProtocol)


class Avilla:
    broadcast: Broadcast
    launch_manager: Launart
    protocols: list[BaseProtocol]
    accounts: dict[Selector, AccountInfo]
    service: AvillaService
    global_artifacts: dict[Any, Any]

    def __init__(
        self,
        *,
        broadcast: Broadcast | None = None,
        launch_manager: Launart | None = None,
        message_cache_size: int = 300,
        record_send: bool = True,
    ):
        self.broadcast = broadcast or it(Broadcast)
        self.launch_manager = launch_manager or it(Launart)
        self.protocols = []
        self._protocol_map = {}
        self.accounts = {}

        self.service = AvillaService(self, message_cache_size)
        self.global_artifacts = {}

        self.launch_manager.add_component(MemcacheService())
        self.launch_manager.add_component(self.service)
        self.broadcast.finale_dispatchers.append(AvillaBuiltinDispatcher(self))

        self.__init_isolate__()

        if message_cache_size > 0:
            from avilla.core.context import Context
            from avilla.core.message import Message
            from avilla.standard.core.account import AccountUnregistered
            from avilla.standard.core.message import MessageReceived

            @self.broadcast.receiver(MessageReceived)
            async def message_cacher(context: Context, message: Message):
                if context.account.info.enabled_message_cache:
                    self.service.message_cache[context.account.route].push(message)

            @self.broadcast.receiver(AccountUnregistered)
            async def clear_cache(event: AccountUnregistered):
                if event.account.route in self.service.message_cache:
                    del self.service.message_cache[event.account.route]

            message_cacher.__annotations__ = {"context": Context, "message": Message}
            clear_cache.__annotations__ = {"event": AccountUnregistered}

        if record_send:
            from avilla.core.context import Context
            from avilla.core.message import Message
            from avilla.standard.core.message import MessageSent

            @self.broadcast.receiver(MessageSent)
            async def message_sender(context: Context, message: Message):
                scene = f"{'.'.join(f'{k}({v})' for k, v in context.scene.items())}"
                logger.info(
                    f"[{context.account.info.protocol.__class__.__name__.replace('Protocol', '')} "
                    f"{context.account.route['account']}]: "
                    f"{scene} <- {str(message.content)!r}"
                )

            message_sender.__annotations__ = {"context": Context, "message": Message}

        self.custom_event_recorder: dict[type[AvillaEvent], Callable[[AvillaEvent], None]] = {}

    def event_record(self, event: AvillaEvent | AvillaLifecycleEvent):
        from avilla.standard.core.account.event import AccountStatusChanged
        from avilla.standard.core.application import AvillaLifecycleEvent
        from avilla.standard.core.message import (
            MessageEdited,
            MessageReceived,
            MessageSent,
        )

        if isinstance(event, AccountStatusChanged):
            logger.debug(
                f"[{event.account.info.protocol.__class__.__name__.replace('Protocol', '')} "
                f"{event.account.route['account']}]: "
                f"{event.__class__.__name__}"
            )
            return
        if isinstance(event, AvillaLifecycleEvent):
            logger.debug(event.__class__.__name__)
            return
        context = event.context
        client = context.client.display
        scene = context.scene.display
        endpoint = context.endpoint.display

        if isinstance(event, MessageSent):
            return

        if type(event) in self.custom_event_recorder:
            self.custom_event_recorder[type(event)](event)

        elif isinstance(event, MessageReceived):
            logger.info(
                f"[{context.account.info.protocol.__class__.__name__.replace('Protocol', '')} "
                f"{context.account.route['account']}]: "
                f"{scene} -> {str(event.message.content)!r}"
            )
        elif isinstance(event, MessageEdited):
            logger.info(
                f"[{context.account.info.protocol.__class__.__name__.replace('Protocol', '')} "
                f"{context.account.route['account']}]: "
                f"{scene} => {str(event.past)!r} -> {str(event.current)!r}"
            )
        else:
            logger.info(
                f"[{context.account.info.protocol.__class__.__name__.replace('Protocol', '')} "
                f"{context.account.route['account']}]: "
                f"{event.__class__.__name__} from {client} to {endpoint} in {scene}"
            )

    @overload
    def add_event_recorder(self, event_type: type[TE]) -> Callable[[Callable[[TE], None]], Callable[[TE], None]]:
        ...

    @overload
    def add_event_recorder(self, event_type: type[TE], recorder: Callable[[TE], None]) -> Callable[[TE], None]:
        ...

    def add_event_recorder(
        self, event_type: type[TE], recorder: Callable[[TE], None] | None = None
    ) -> Callable[..., Callable[[TE], None]] | Callable[[TE], None]:
        if recorder is None:

            def wrapper(func: Callable[[TE], None]):
                self.custom_event_recorder[event_type] = func  # type: ignore
                return func

            return wrapper
        self.custom_event_recorder[event_type] = recorder  # type: ignore
        return recorder

    def __init_isolate__(self):
        from avilla.core.builtins.resource_fetch import CoreResourceFetchPerform

        CoreResourceFetchPerform.apply_to(self.global_artifacts)

    def get_staff_components(self):
        return {"avilla": self}

    def get_staff_artifacts(self):
        return [self.global_artifacts]

    def __staff_generic__(self, element_type: dict, event_type: dict):
        ...

    @classmethod
    def current(cls):
        return get_current_avilla()

    async def fetch_resource(self, resource: Resource[T]) -> T:
        return await Staff(self.get_staff_artifacts(), self.get_staff_components()).fetch_resource(resource)

    def get_account(self, target: Selector) -> AccountInfo:
        return self.accounts[target]

    @overload
    def get_accounts(self, *, land: str) -> list[AccountInfo]:
        ...

    @overload
    def get_accounts(self, *, pattern: str) -> list[AccountInfo]:
        ...

    @overload
    def get_accounts(self, *, protocol_type: type[BaseProtocol]) -> list[AccountInfo]:
        ...

    @overload
    def get_accounts(self, *, account_type: type[BaseAccount]) -> list[AccountInfo]:
        ...

    def get_accounts(
        self,
        *,
        land: str | None = None,
        pattern: str | None = None,
        protocol_type: type[BaseProtocol] | None = None,
        account_type: type[BaseAccount] | None = None,
    ) -> list[AccountInfo]:
        if land:
            return [account for account in self.accounts.values() if account.platform.land.name == land]
        if pattern:
            return [account for selector, account in self.accounts.items() if selector.follows(pattern)]
        if protocol_type:
            return [account for account in self.accounts.values() if isinstance(account.protocol, protocol_type)]
        if account_type:
            return [account for account in self.accounts.values() if isinstance(account.account, account_type)]
        raise ValueError("No filter is specified")

    def _update_protocol_map(self):
        self._protocol_map = {type(i): i for i in self.protocols}

    def apply_protocols(self, *protocols: BaseProtocol) -> None:
        self.protocols.extend(protocols)
        self._update_protocol_map()

        for protocol in protocols:
            protocol.ensure(self)

    def get_protocol(self, protocol_type: type[TP]) -> TP:
        if protocol_type not in self._protocol_map:
            raise ValueError(f"{identity(protocol_type)} is unregistered on this Avilla instance")

        return self._protocol_map[protocol_type]  # type: ignore

    def apply_services(self, *services: Service):
        for i in services:
            self.launch_manager.add_component(i)

    def listen(
        self,
        event: type[Dispatchable],
        priority: int = 16,
        dispatchers: list[T_Dispatcher] | None = None,
        namespace: Namespace | None = None,
        decorators: list[Decorator] | None = None,
    ):
        return self.broadcast.receiver(event, priority, dispatchers, namespace, decorators)

    def launch(
        self,
        *,
        loop: asyncio.AbstractEventLoop | None = None,
        stop_signal: Iterable[signal.Signals] = (signal.SIGINT,),
    ):
        self.launch_manager.launch_blocking(loop=loop, stop_signal=stop_signal)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Awaitable, Callable, TypeVar, overload

from graia.ryanvk import BaseCollector

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from avilla.core.metadata import Metadata, MetadataRoute
    from avilla.core.selector import Selector

T = TypeVar("T")
T1 = TypeVar("T1")

M = TypeVar("M", bound="Metadata")


Wrapper = Callable[[T], T]


class AvillaBaseCollector(BaseCollector):
    def __init__(self):
        super().__init__()

    @overload
    def pull(
        self, target: str, route: type[M]
    ) -> Callable[[Callable[[Any, Selector, type[M]], Awaitable[M]]], Callable[[Any, Selector, type[M]], Awaitable[M]]]:
        ...

    @overload
    def pull(
        self, target: str, route: MetadataRoute[Unpack[tuple[Any, ...]], M]
    ) -> Callable[[Callable[[Any, Selector, type[M]], Awaitable[M]]], Callable[[Any, Selector, type[M]], Awaitable[M]]]:
        ...

    def pull(self, target: str, route: ...) -> ...:
        from avilla.core.builtins.capability import CoreCapability

        return self.entity(CoreCapability.pull, target=target, route=route)

    def fetch(self, resource_type: type[T]) -> Wrapper[Callable[[Any, T], Awaitable[Any]]]:
        from avilla.core.builtins.capability import CoreCapability

        return self.entity(CoreCapability.fetch, resource=resource_type)  # type: ignore



from .event import AccountAvailable as AccountAvailable
from .event import AccountRegistered as AccountRegistered
from .event import AccountStatusChanged as AccountStatusChanged
from .event import AccountUnavailable as AccountUnavailable
from .event import AccountUnregistered as AccountUnregistered



from __future__ import annotations

from pathlib import Path
from typing import Generic, TypeVar

from avilla.core.selector import Selector

from .metadata import Metadata

T = TypeVar("T")


class Resource(Metadata, Generic[T]):
    selector: Selector

    def __init__(self, selector: Selector):
        self.selector = selector

    def to_selector(self):
        return self.selector


class LocalFileResource(Resource[bytes]):
    file: Path

    def __init__(self, file: Path | str):
        if isinstance(file, str):
            file = Path(file)
        self.file = file

    @property
    def selector(self) -> Selector:
        return Selector().land("avilla-core").local_file(str(self.file))


class RawResource(Resource[T]):
    data: T

    def __init__(self, data: T):
        self.data = data

    @property
    def selector(self):
        return Selector().land("avilla-core").raw_data(str(id(self)))


class UrlResource(Resource[bytes]):
    def __init__(self, url: str):
        self.url = url

    @property
    def selector(self):
        return Selector().land("avilla-core").url(self.url)



from __future__ import annotations

from contextlib import AsyncExitStack, asynccontextmanager
from typing import TYPE_CHECKING, Any, ClassVar

from .endpoint import Endpoint

if TYPE_CHECKING:
    from graia.ryanvk.staff import Staff

    from .collector import BaseCollector


class BasePerform:
    __collector__: ClassVar[BaseCollector]
    # spec said one perform declare / class only binds to one collector instance.
    # multi to single or reversed behavior or settings are both denied in spec,
    # and, be suggested, actually coding. 

    __static__: ClassVar[bool] = True
    # when a perform is static, its lifespan won't execute,
    # which means dynamic endpoint cannot be used in the perform.

    staff: Staff

    def __init__(self, staff: Staff) -> None:
        self.staff = staff

    def __post_init__(self):
        ...

    @classmethod
    def apply_to(cls, map: dict[Any, Any]):
        map.update(cls.__collector__.artifacts)

    @classmethod
    def endpoints(cls):
        return [(k, v) for k, v in cls.__dict__.items() if isinstance(v, Endpoint)]

    @asynccontextmanager
    async def lifespan(self):
        async with AsyncExitStack() as stack:
            for _, v in self.endpoints():
                await stack.enter_async_context(v.lifespan(self))
            yield self

    @classmethod
    def __post_collected__(cls, collect: BaseCollector):
        ...

    def __init_subclass__(cls, *, native: bool = False, static: bool = True) -> None:
        if native:
            return

        collector = cls.__collector__
        cls.__static__ = static

        for i in collector.collected_callbacks:
            i(cls)

        cls.__post_collected__(collector)



from __future__ import annotations

import functools
from typing import TYPE_CHECKING

from graia.broadcast.utilles import Ctx

if TYPE_CHECKING:
    from avilla.core.application import Avilla
    from avilla.core.context import Context
    from avilla.core.protocol import BaseProtocol


cx_avilla: Ctx[Avilla] = Ctx("avilla")
cx_protocol: Ctx[BaseProtocol] = Ctx("protocol")
cx_context: Ctx[Context] = Ctx("context")


def get_current_avilla() -> Avilla:
    avilla = cx_avilla.get(None)
    if avilla:
        return avilla
    protocol = cx_protocol.get(None)
    if protocol:
        return protocol.avilla
    context = cx_context.get(None)
    if context:
        return context.protocol.avilla
    raise RuntimeError("no any current avilla")


def get_current_protocol():
    protocol = cx_protocol.get(None)
    if protocol:
        return protocol
    context = cx_context.get(None)
    if context:
        return context.protocol
    raise RuntimeError("no any current protocol")


def require_context(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        if cx_context.get(None):
            return await func(*args, **kwargs)
        raise RuntimeError("no any current context")

    return wrapper



from __future__ import annotations

from typing import Any, Callable, TypedDict

from avilla.core.ryanvk.descriptor.query import QueryRecord, find_querier_steps
from avilla.core.selector import FollowsPredicater, _parse_follows
from graia.ryanvk import FnOverload, OverloadBehavior
from graia.ryanvk.collector import BaseCollector


class QueryCollectParams(TypedDict):
    target: str
    previous: str | None


class QueryCallArgs(TypedDict):
    pattern: str
    predicators: dict[str, FollowsPredicater]


class QueryOverload(FnOverload):
    identity: str = "query"

    def collect_entity(
        self,
        collector: BaseCollector,
        scope: dict[Any, Any],
        entity: Any,
        params: QueryCollectParams,
    ) -> None:
        record_tuple = (collector, entity)

        sign = QueryRecord(params["previous"], params["target"])
        scope.setdefault(sign, set()).add(record_tuple)

    def get_entities(
        self,
        scope: dict[QueryRecord, set[tuple[BaseCollector, Callable[..., Any]]]],
        args: QueryCallArgs,
    ):
        items = _parse_follows(args["pattern"], **args["predicators"])
        steps = find_querier_steps(scope, items)

        if steps is None:
            raise NotImplementedError

        ...


class QueryFnBehavior(OverloadBehavior):
    ...



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TypedDict, TypeVar


@dataclass
class PlatformDescription:
    ...


PD = TypeVar("PD", bound=PlatformDescription)


@dataclass
class Platform:
    description: dict[type[PlatformDescription], PlatformDescription]

    def __init__(self, land: Land, abstract: Abstract, *description: PlatformDescription) -> None:
        self.description = {type(i): i for i in description}
        self.description[Land] = land
        self.description[Abstract] = abstract

    def __getitem__(self, item: type[PD]) -> PD:
        return self.description[item]  # type: ignore

    def __str__(self):
        return f"<Platform {' '.join([str(i) for i in self.description.values()])}>"

    @property
    def land(self):
        return self[Land]


class Maintainer(TypedDict):
    name: str


@dataclass
class Land(PlatformDescription):
    name: str
    maintainers: list[Maintainer] = field(default_factory=list)
    humanized_name: str | None = None


@dataclass
class Abstract(PlatformDescription):
    protocol: str
    maintainers: list[Maintainer] = field(default_factory=list)
    humanized_name: str | None = None


@dataclass
class Distribution(PendingDeprecationWarning):
    name: str
    base: str
    maintainers: list[Maintainer] = field(default_factory=list)
    humanized_name: str | None = None


@dataclass
class Branch(PlatformDescription):
    value: str


@dataclass
class Version(PlatformDescription):
    components: dict[str, str]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

from typing_extensions import Unpack

from avilla.core.metadata import Metadata, MetadataRoute
from avilla.core.resource import Resource, T
from avilla.core.ryanvk.descriptor.query import QuerySchema
from avilla.core.ryanvk.overload.metadata import MetadataOverload
from avilla.core.ryanvk.overload.target import TargetOverload
from graia.ryanvk.capability import Capability
from graia.ryanvk.fn import Fn
from graia.ryanvk.overload import NoneOverload, TypeOverload

if TYPE_CHECKING:
    from avilla.core.context import Context
    from avilla.core.selector import Selector


M = TypeVar("M", bound=Metadata)


class CoreCapability(Capability):
    query = QuerySchema()

    @Fn.complex(
        {
            TargetOverload(): ["target"],
            NoneOverload(TargetOverload(), default_factory=lambda _: None): ["via"],
        }
    )
    def get_context(self, target: Selector, *, via: Selector | None = None) -> Context:
        ...

    @Fn.complex({TargetOverload(): ["target"], MetadataOverload(): ["route"]})
    async def pull(self, target: Selector, route: type[M] | MetadataRoute[Unpack[tuple[Any, ...]], M]) -> Any:
        ...

    @Fn.complex({TypeOverload(): ["resource"]})
    async def fetch(self, resource: Resource[T]) -> T:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    def channel(self, target: Selector) -> str:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    def guild(self, target: Selector) -> str:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    def user(self, target: Selector) -> str:
        ...



from __future__ import annotations

from avilla.core.ryanvk import Capability, Fn, TargetOverload
from avilla.core.selector import Selector


class RequestCapability(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def accept(self, target: Selector):
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def reject(self, target: Selector, reason: str | None = None, forever: bool = False):
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def cancel(self, target: Selector):
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def ignore(self, target: Selector):
        ...



from __future__ import annotations

from dataclasses import dataclass

from graia.broadcast.interfaces.dispatcher import DispatcherInterface

from avilla.core.event import AvillaEvent
from avilla.core.request import Request


@dataclass
class RequestEvent(AvillaEvent):
    request: Request

    class Dispatcher(AvillaEvent.Dispatcher):
        @staticmethod
        async def catch(interface: DispatcherInterface[RequestEvent]):
            if interface.annotation is Request:
                return interface.event.request
            if interface.name == "sender":
                return interface.event.request.sender
            return await AvillaEvent.Dispatcher.catch(interface)


class RequestReceived(RequestEvent):
    pass


class RequestAccepted(RequestEvent):
    pass


class RequestRejected(RequestEvent):
    pass


class RequestIgnored(RequestEvent):
    pass


class RequestCancelled(RequestEvent):
    pass



from typing import Any, Protocol, TypeVar, runtime_checkable

# if TYPE_CHECKING:
#     pass

"""
TProtocol = TypeVar("TProtocol", bound="BaseProtocol")


ActionExtensionImpl = Callable[
    ["RelationshipExecutor", "ActionExtension", dict[str, Any] | None], Coroutine[None, None, Any]
]
"""

_T = TypeVar("_T", contravariant=True)


@runtime_checkable
class Ensureable(Protocol[_T]):
    def ensure(self, interact: _T) -> Any:
        ...



from __future__ import annotations

from dataclasses import dataclass

from avilla.core.metadata import Metadata


@dataclass
class Reason(Metadata):
    content: str


@dataclass
class Comment(Metadata):
    content: str


@dataclass
class QuestionItem:
    id: str
    content: str


@dataclass
class Questions(Metadata):
    items: list[QuestionItem]


@dataclass
class AnswerItem:
    id: str
    content: str


@dataclass
class Answers(Metadata):
    items: list[AnswerItem]



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, TypeVar

from graia.ryanvk import Access, BasePerform

from .base import AvillaBaseCollector

if TYPE_CHECKING:
    from avilla.core.application import Avilla


T = TypeVar("T")
T1 = TypeVar("T1")


class ApplicationBasedPerformTemplate(BasePerform, native=True):
    __collector__: ClassVar[ApplicationCollector]

    avilla: Access[Avilla] = Access()


class ApplicationCollector(AvillaBaseCollector):
    post_applying: bool = False

    def __init__(self):
        super().__init__()

    @property
    def _(self):
        upper = super()._

        class LocalPerformTemplate(
            ApplicationBasedPerformTemplate,
            upper,
            native=True,
        ):
            ...

        return LocalPerformTemplate



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable

from typing_extensions import TypeAlias

from avilla.core.selector import FollowsPredicater, Selector, _parse_follows
from graia.ryanvk.collector import BaseCollector
from graia.ryanvk.overload import FnOverload


@dataclass
class LookupBranchMetadata:
    ...


@dataclass
class LookupBranch:
    metadata: LookupBranchMetadata
    levels: LookupCollection
    bind: set[tuple[BaseCollector, Callable]] = field(default_factory=set)


LookupBranches: TypeAlias = "dict[str | FollowsPredicater | None, LookupBranch]"
LookupCollection: TypeAlias = "dict[str, LookupBranches]"


@dataclass
class TargetOverloadConfig:
    pattern: str
    predicators: dict[str, FollowsPredicater]

    def __init__(self, pattern: str, **predicators: FollowsPredicater):
        self.pattern = pattern
        self.predicators = predicators


def _merge_lookup_collection(current: LookupCollection, other: LookupCollection):
    for key, branches in current.items():
        if (other_branches := other.pop(key, None)) is None:
            continue

        for header, branch in branches.items():
            if (other_branch := other_branches.pop(header, None)) is None:
                continue

            _merge_lookup_collection(branch.levels, other_branch.levels)
            branch.bind |= other_branch.bind

        branches |= other_branches

    current |= other


class TargetOverload(FnOverload):
    def collect_entity(
        self,
        collector: BaseCollector,
        scope: dict[Any, Any],
        entity: Any,
        params: dict[str, str | TargetOverloadConfig],
    ) -> None:
        record = (collector, entity)

        for param, pattern in params.items():
            param_scope = scope.setdefault(param, {})

            if isinstance(pattern, str):
                pattern = TargetOverloadConfig(pattern)

            pattern_items = _parse_follows(pattern.pattern, **pattern.predicators)
            if not pattern_items:
                raise ValueError("invalid target pattern")

            processing_level = param_scope

            if TYPE_CHECKING:
                branch = LookupBranch(LookupBranchMetadata(), {})

            for item in pattern_items:
                if item.name not in processing_level:
                    processing_level[item.name] = {}

                branches = processing_level[item.name]
                if (item.literal or item.predicate) in branches:
                    branch = branches[item.literal or item.predicate]
                else:
                    branch = LookupBranch(LookupBranchMetadata(), {})
                    branches[item.literal or item.predicate] = branch

                processing_level = branch.levels

            branch.bind.add(record)

    def get_entities(self, scope: dict[Any, Any], args: dict[str, Selector]) -> set[tuple[BaseCollector, Callable]]:
        bind_sets: list[set] = []

        for arg_name, selector in args.items():
            if arg_name not in scope:
                raise NotImplementedError

            def get_bind_set():
                processing_scope: LookupCollection = scope[arg_name]
                branch = None
                for key, value in selector.pattern.items():
                    if (branches := processing_scope.get(key)) is None:
                        raise NotImplementedError

                    if value in branches:
                        header = value
                    else:
                        for _key, branch in branches.items():
                            if callable(_key) and _key(value):
                                header = _key
                                break  # hit predicate
                        else:
                            if None in branches:
                                header = None  # hit default
                            elif "*" in branches:
                                return branches["*"].bind  # hit wildcard
                            else:
                                raise NotImplementedError

                    branch = branches[header]
                    processing_scope = branch.levels

                    if header is not None and None in branches:
                        processing_scope = branches[None].levels | processing_scope
                if branch is not None and branch.bind:
                    # branch has bind
                    return branch.bind

                raise NotImplementedError

            bind_sets.append(get_bind_set())
        return bind_sets.pop().intersection(*bind_sets)

    def merge_scopes(self, *scopes: dict[Any, Any]):
        # scope layout: {
        #   <param_name: str>: LookupCollection
        # }
        param_collections: dict[str, list[LookupCollection]] = {}
        result = {}

        for scope in scopes:
            for param, collection in scope.items():
                param_collections.setdefault(param, []).append(collection)

        for param, collections in param_collections.items():
            result[param] = current = collections.pop(0)
            for other in collections:
                _merge_lookup_collection(current, other)

        return result



from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeVar

from graia.amnesia.message import Element, MessageChain, Text
from typing_extensions import ParamSpec

from avilla.core.builtins.capability import CoreCapability
from avilla.core.message import Message
from avilla.core.metadata import Metadata
from avilla.core.selector import Selector
from avilla.standard.core.activity import ActivityTrigger
from avilla.standard.core.message import MessageSend
from avilla.standard.core.relation import SceneCapability
from avilla.standard.core.request import RequestCapability

from ._selector import ContextSelector

if TYPE_CHECKING:
    pass

R = TypeVar("R", covariant=True)
P = ParamSpec("P")
_MetadataT = TypeVar("_MetadataT", bound=Metadata)


class ContextClientSelector(ContextSelector):
    def trigger_activity(self, activity: str):
        return self.context[ActivityTrigger.trigger](self.activity(activity))

    @property
    def channel(self) -> str:
        return self.context.staff.call_fn(CoreCapability.channel, self)

    @property
    def guild(self) -> str | None:
        try:
            return self.context.staff.call_fn(CoreCapability.guild, self)
        except NotImplementedError:
            return None

    @property
    def user(self) -> str:
        return self.context.staff.call_fn(CoreCapability.user, self)


class ContextEndpointSelector(ContextSelector):
    def expects_request(self) -> ContextRequestSelector:
        if "request" in self.pattern:
            return ContextRequestSelector.from_selector(self.context, self)

        raise ValueError(f"endpoint {self!r} is not a request endpoint")

    @property
    def channel(self) -> str:
        return self.context.staff.call_fn(CoreCapability.channel, self)

    @property
    def guild(self) -> str | None:
        try:
            return self.context.staff.call_fn(CoreCapability.guild, self)
        except NotImplementedError:
            return None

    @property
    def user(self) -> str:
        return self.context.staff.call_fn(CoreCapability.user, self)


class ContextSelfSelector(ContextSelector):
    def trigger_activity(self, activity: str):
        return self.context[ActivityTrigger.trigger](self.activity(activity))


class ContextSceneSelector(ContextSelector):
    def leave_scene(self):
        return self.context[SceneCapability.leave](self)

    def disband_scene(self):
        return self.context[SceneCapability.disband](self)

    def send_message(
        self,
        message: MessageChain | Iterable[str | Element] | Element | str,
        *,
        reply: Message | Selector | str | None = None,
    ):
        if isinstance(message, str):
            message = MessageChain([Text(message)])
        elif isinstance(message, Element):
            message = MessageChain([message])
        elif not isinstance(message, MessageChain):
            message = MessageChain([]).extend(list(message))
        else:
            message = MessageChain([i if isinstance(i, Element) else Text(i) for i in message])

        if isinstance(reply, Message):
            reply = reply.to_selector()
        elif isinstance(reply, str):
            reply = self.message(reply)

        return self.context[MessageSend.send](self, message, reply=reply)

    def remove_member(self, target: Selector, reason: str | None = None):
        return self.context[SceneCapability.remove_member](target, reason)

    @property
    def channel(self) -> str:
        return self.context.staff.call_fn(CoreCapability.channel, self)

    @property
    def guild(self) -> str:
        return self.context.staff.call_fn(CoreCapability.guild, self)


class ContextRequestSelector(ContextEndpointSelector):
    def accept_request(self):
        return self.context[RequestCapability.accept](self)

    def reject_request(self, reason: str | None = None, forever: bool = False):
        return self.context[RequestCapability.reject](self, reason, forever)

    def cancel_request(self):
        return self.context[RequestCapability.cancel](self)

    def ignore_request(self):
        return self.context[RequestCapability.ignore](self)


@dataclass
class ContextMedium:
    selector: ContextSelector



from __future__ import annotations

from collections.abc import Awaitable, Callable, Container, Iterable
from functools import partial
from inspect import isawaitable
from operator import contains, eq, is_, is_not, ne
from typing import Any, Generic, Protocol, TypeVar

from graia.broadcast import BaseDispatcher, DispatcherInterface, ExecutionStop
from typing_extensions import ParamSpec, Self

from avilla.core.account import BaseAccount
from avilla.core.context import Context
from avilla.core.event import MetadataModified
from avilla.core.selector import Selectable, Selector
from avilla.core.utilles import classproperty

T = TypeVar("T", covariant=True)
R = TypeVar("R")
P = ParamSpec("P")


class _ClientSpecified(Protocol[R]):
    client: R


class _EndpointSpecified(Protocol[R]):
    endpoint: R


class _SceneSpecified(Protocol[R]):
    scene: R


class _SelfSpecified(Protocol[R]):
    self: R


class Filter(BaseDispatcher, Generic[T]):
    middlewares: list[Callable[[Any], Awaitable[Any] | Any]]

    def __init__(self):
        self.middlewares = []

    def dispatch(self, annotation: type[R]) -> Filter[R]:
        @self.middlewares.append
        async def _(_) -> R:
            interface = DispatcherInterface.ctx.get()
            result = await interface.lookup_param("__filter_fetch__", annotation, None)
            interface.local_storage.setdefault("__filter_fetch__", {})[annotation] = result
            return result

        return self  # type: ignore

    def step(self, func: Callable[[T], R]) -> Filter[R]:
        self.middlewares.append(func)
        return self  # type: ignore

    def assert_true(self: Filter[R], func: Callable[[R], Any]):
        @self.step
        def _(result: R) -> R:
            if func(result):
                return result
            raise ExecutionStop

        return self

    def assert_false(self, func: Callable[[T], Any]) -> Self:
        return self.assert_true(lambda result: not func(result))

    def assert_equal(self, value: Any) -> Self:
        return self.assert_true(partial(eq, value))

    def assert_not_equal(self, value: Any) -> Self:
        return self.assert_true(partial(ne, value))

    def is_(self, value: Any):
        return self.assert_true(partial(is_, value))

    def is_not(self, value: Any) -> Self:
        return self.assert_true(partial(is_not, value))

    def is_none(self):
        return self.is_(None)

    def is_not_none(self) -> Self:  # 实际上类型是 Filter[T - None]
        return self.is_not(None)

    def contains(self, container: Container[R]) -> Filter[R]:
        return self.assert_true(partial(contains, container))  # type: ignore

    def not_contains(self, container: Container) -> Self:
        return self.assert_true(lambda result: result not in container)

    def instanceof(self, cls: type[R]) -> Filter[R]:
        return self.assert_true(lambda result: isinstance(result, cls))  # type: ignore

    def not_instanceof(self, cls: type) -> Self:
        return self.assert_true(lambda result: not isinstance(result, cls))

    def any(self, funcs: Iterable[Callable[[T], Any]]) -> Self:
        funcs = frozenset(funcs)
        return self.assert_true(lambda result: any(func(result) for func in funcs))

    def all(self, funcs: Iterable[Callable[[T], Any]]) -> Self:
        funcs = frozenset(funcs)
        return self.assert_true(lambda result: all(func(result) for func in funcs))

    def follows(self: Filter[Selectable], *patterns: str) -> Filter[Selectable]:
        return self.any(lambda result: result.to_selector().follows(pattern) for pattern in patterns)

    @classproperty
    @classmethod
    def cx(cls) -> Filter[Context]:
        return cls().dispatch(Context)

    @property
    def client(self: Filter[_ClientSpecified[R]]) -> Filter[R]:
        return self.step(lambda x: x.client)

    @property
    def scene(self: Filter[_SceneSpecified[R]]) -> Filter[R]:
        return self.step(lambda x: x.scene)

    @property
    def endpoint(self: Filter[_EndpointSpecified[R]]) -> Filter[R]:
        return self.step(lambda x: x.endpoint)

    @property
    def self(self: Filter[_SelfSpecified[R]]) -> Filter[R]:
        return self.step(lambda x: x.self)

    @classmethod
    def account(cls) -> Filter[BaseAccount]:
        return cls.cx.step(lambda cx: cx.account)

    def medium(self: Filter[Context], index: int = 0) -> Filter[Selector | None]:
        return self.cx.step(lambda cx: cx.mediums[index].selector if cx.mediums else None)

    @classproperty
    @classmethod
    def mod(cls) -> Filter[MetadataModified]:
        return cls().dispatch(MetadataModified)

    async def beforeExecution(self, interface: DispatcherInterface):
        result: Awaitable[Any] | Any = interface  # type: ignore
        for middleware in self.middlewares:
            if isawaitable(result := middleware(result)):
                result = await result

    async def catch(self, interface: DispatcherInterface):
        return interface.local_storage.get("__filter_fetch__", {}).get(interface.annotation)


if __name__ == "__main__":
    print(Filter().scene.follows("group").middlewares)



from __future__ import annotations

from collections import deque
from dataclasses import dataclass
from typing import Any, AsyncGenerator, Callable, Container, Protocol, overload

from avilla.core.selector import Selector, _FollowItem
from graia.ryanvk import BaseCollector


@dataclass(unsafe_hash=True)
class QueryRecord:
    """仅用作计算路径, 不参与实际运算, 也因此, 该元素仅存在于全局 Artifacts['query'] 中."""

    previous: str | None
    into: str


class QueryHandlerPerform(Protocol):
    def __call__(
        self, _p0: Any, predicate: Callable[[str, str], bool] | str, previous: Selector | None = None
    ) -> AsyncGenerator[Selector, None]:
        ...


class QueryHandlerPerformNoPrev(Protocol):
    def __call__(
        self, _p0: Any, predicate: Callable[[str, str], bool] | str, previous: None
    ) -> AsyncGenerator[Selector, None]:
        ...


class QueryHandlerPerformPrev(Protocol):
    def __call__(
        self, _p0: Any, predicate: Callable[[str, str], bool] | str, previous: Selector
    ) -> AsyncGenerator[Selector, None]:
        ...


class QuerySchema:
    @overload
    def collect(
        self, collector: BaseCollector, target: str, previous: None = None
    ) -> Callable[[QueryHandlerPerformNoPrev], QueryHandlerPerformNoPrev]:
        ...

    @overload
    def collect(
        self, collector: BaseCollector, target: str, previous: str
    ) -> Callable[[QueryHandlerPerformPrev], QueryHandlerPerformPrev]:
        ...

    def collect(self, collector: BaseCollector, target: str, previous: ... = None) -> ...:
        def receive(entity: QueryHandlerPerform):
            collector.artifacts[QueryRecord(previous, target)] = (collector, entity)
            return entity

        return receive


class QueryHandler(Protocol):
    def __call__(
        self, predicate: Callable[[str, str], bool] | str, previous: Selector | None = None
    ) -> AsyncGenerator[Selector, None]:
        ...


# 使用 functools.reduce.
async def query_depth_generator(
    handler: QueryHandler,
    predicate: Callable[[str, str], bool] | str,
    previous_generator: AsyncGenerator[Selector, None] | None = None,
):
    if previous_generator is not None:
        async for previous in previous_generator:
            async for current in handler(predicate, previous):
                yield current
    else:
        async for current in handler(predicate):
            yield current


@dataclass
class _MatchStep:
    upper: str
    start: int
    history: tuple[tuple[tuple[_FollowItem, ...], QueryRecord], ...]


def find_querier_steps(
    artifacts: Container[Any],
    frags: list[_FollowItem],
) -> list[tuple[tuple[_FollowItem, ...], QueryRecord]] | None:
    result: list[tuple[tuple[_FollowItem, ...], QueryRecord]] | None = None
    queue: deque[_MatchStep] = deque([_MatchStep("", 0, ())])
    whole = ".".join([i.name for i in frags])
    while queue:
        head: _MatchStep = queue.popleft()
        current_steps: list[_FollowItem] = []
        for curr_frag in frags[head.start :]:
            current_steps.append(curr_frag)
            steps = ".".join([i.name for i in current_steps])
            full_path = f"{head.upper}.{steps}" if head.upper else steps
            head.start += 1
            if (query := ((*current_steps,), QueryRecord(head.upper or None, steps)))[1] in artifacts:
                if full_path == whole:
                    if result is None or len(result) > len(head.history) + 1:
                        result = [*head.history, query]
                else:
                    queue.append(_MatchStep(full_path, head.start, head.history + (query,)))
    return result



from __future__ import annotations

from typing import Any, Callable

from graia.ryanvk.collector import BaseCollector


class FnOverload:
    @property
    def identity(self) -> str:
        return self.__class__.__name__

    def collect_entity(
        self,
        collector: BaseCollector,
        scope: dict[Any, Any],
        entity: Any,
        params: dict[str, Any],
    ) -> None:
        scope["_"] = {(collector, entity)}

    def get_entities(self, scope: dict[Any, Any], args: dict[str, Any]) -> set[tuple[BaseCollector, Callable]]:
        return scope["_"]

    def merge_scopes(self, *scopes: dict[Any, Any]) -> dict:
        return scopes[-1]

    def get_params_layout(self, params: list[str], args: dict[str, Any]) -> dict[str, Any]:
        return {param: args[param] for param in params}


class SimpleOverload(FnOverload):
    @property
    def identity(self) -> str:
        return str(id(self))

    def collect_entity(
        self,
        collector: BaseCollector,
        scope: dict[Any, Any],
        entity: Any,
        params: dict[str, Any],
    ) -> None:
        record = (collector, entity)

        for param_name, collect_info in params.items():
            collection = scope.setdefault(param_name, {})
            collection.setdefault(collect_info, set()).add(record)

    def get_entities(self, scope: dict[Any, Any], args: dict[str, Any]) -> set[tuple[BaseCollector, Callable]]:
        result_sets: list[set] = []

        for arg_name, arg_value in args.items():
            collection = scope[arg_name]
            if arg_value not in collection:
                raise NotImplementedError
            result_sets.append(collection[arg_value])

        return result_sets.pop().intersection(*result_sets)

    def merge_scopes(self, *scopes: dict[Any, Any]):
        # layout: {arg: {value: set}}

        result = {}

        for scope in scopes:
            for param_name, param_collection in scope.items():
                collection = result.setdefault(param_name, {})
                for value, entities in param_collection.items():
                    collection.setdefault(value, set()).update(entities)

        return result


class TypeOverload(FnOverload):
    @property
    def identity(self) -> str:
        return "type_overload:" + str(id(self))

    def collect_entity(
        self,
        collector: BaseCollector,
        scope: dict[Any, Any],
        entity: Any,
        params: dict[str, Any],
    ) -> None:
        record = (collector, entity)

        for param_name, collect_info in params.items():
            collection = scope.setdefault(param_name, {})
            collection.setdefault(collect_info, set()).add(record)

    def get_entities(self, scope: dict[Any, Any], args: dict[str, Any]) -> set[tuple[BaseCollector, Callable]]:
        result_sets: list[set] = []

        for arg_name, arg_value in args.items():
            collection = scope[arg_name]
            key = type(arg_value)
            if key not in collection:
                raise NotImplementedError
            result_sets.append(collection[key])

        return result_sets.pop().intersection(*result_sets)

    def merge_scopes(self, *scopes: dict[Any, Any]):
        # layout: {arg: {value: set}}

        result = {}

        for scope in scopes:
            for param_name, param_collection in scope.items():
                collection = result.setdefault(param_name, {})
                for value, entities in param_collection.items():
                    collection.setdefault(value, set()).update(entities)

        return result


class NoneOverload(FnOverload):
    default_factory: Callable[[str], Any] | None = None
    bypassing: FnOverload

    def __init__(self, bypassing: FnOverload, *, default_factory: Callable[[str], Any] | None = None) -> None:
        self.bypassing = bypassing
        self.default_factory = default_factory

    @property
    def identity(self) -> str:
        return "none_overload:" + str(id(self))

    def get_params_layout(self, params: list[str], args: dict[str, Any]) -> dict[str, Any]:
        if self.default_factory is not None:
            return {param: args.get(param) or self.default_factory(param) for param in params}
        return super().get_params_layout(params, args)

    def collect_entity(
        self,
        collector: BaseCollector,
        scope: dict[Any, Any],
        entity: Any,
        params: dict[str, Any],
    ) -> None:
        record = (collector, entity)
        if not scope:
            scope.update({"none": {}, "bypassing": {}})
        for param, collect_info in params.items():
            if collect_info is None:
                scope["none"].setdefault(param, set()).add(record)
            else:
                self.bypassing.collect_entity(
                    collector,
                    scope["bypassing"].setdefault(param, {}),
                    entity,
                    self.bypassing.get_params_layout([param], params),
                )

    def get_entities(
        self, scope: dict[Any, Any], args: dict[str, Any]
    ) -> set[tuple[BaseCollector, Callable[..., Any]]]:
        sets: list[set] = []
        none_collection = scope["none"]
        bypassing_collection = scope["bypassing"]

        for arg_name, arg_value in args.items():
            if arg_value is None:
                sets.append(none_collection[arg_name])
            else:
                sets.append(self.bypassing.get_entities(bypassing_collection[arg_name], {arg_name: arg_value}))

        return sets.pop().intersection(*sets)

class PredicateOverload(FnOverload):
    predicate: Callable[[str, Any], Any]

    def __init__(self, predicate: Callable[[str, Any], Any]) -> None:
        self.predicate = predicate

    @property
    def identity(self) -> str:
        return "predicate_overload:" + str(id(self))

    def collect_entity(
        self,
        collector: BaseCollector,
        scope: dict[Any, Any],
        entity: Any,
        params: dict[str, Any],
    ) -> None:
        record = (collector, entity)

        for param_name, collect_info in params.items():
            collection = scope.setdefault(param_name, {})
            collection.setdefault(collect_info, set()).add(record)

    def get_entities(self, scope: dict[Any, Any], args: dict[str, Any]) -> set[tuple[BaseCollector, Callable]]:
        result_sets: list[set] = []

        for arg_name, arg_value in args.items():
            collection = scope[arg_name]
            key = self.predicate(arg_name, arg_value)
            if key not in collection:
                raise NotImplementedError
            result_sets.append(collection[key])

        return result_sets.pop().intersection(*result_sets)

    def merge_scopes(self, *scopes: dict[Any, Any]):
        # layout: {arg: {value: set}}

        result = {}

        for scope in scopes:
            for param_name, param_collection in scope.items():
                collection = result.setdefault(param_name, {})
                for value, entities in param_collection.items():
                    collection.setdefault(value, set()).update(entities)

        return result



from __future__ import annotations

from collections import ChainMap
from contextlib import AsyncExitStack, asynccontextmanager
from copy import copy
from typing import TYPE_CHECKING, Any, Callable, Protocol, TypeVar, overload

from typing_extensions import ParamSpec

if TYPE_CHECKING:
    from .fn import Fn
    from .perform import BasePerform

P = ParamSpec("P")
R = TypeVar("R", covariant=True)
VnCallable = TypeVar("VnCallable", bound=Callable)

class Staff:
    artifact_collections: list[dict[Any, Any]]
    artifact_map: ChainMap[Any, Any]
    components: dict[str, Any]
    exit_stack: AsyncExitStack
    instances: dict[type, Any]

    def __init__(self, artifacts_collections: list[dict[Any, Any]], components: dict[str, Any]) -> None:
        self.artifact_collections = artifacts_collections
        self.artifact_map = ChainMap(*artifacts_collections)
        self.components = components
        self.exit_stack = AsyncExitStack()
        self.instances = {}

    def call_fn(self, fn: Fn[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
        collector, entity = fn.behavior.harvest_overload(self, fn, *args, **kwargs)
        return fn.execute(self, collector, entity, *args, **kwargs)

    class PostInitShape(Protocol[P]):
        def __post_init__(self, *args: P.args, **kwargs: P.kwargs) -> Any:
            ...

    @overload
    def inject(self, perform_type: type[PostInitShape[P]], *args: P.args, **kwargs: P.kwargs) -> None:
        ...

    @overload
    def inject(self, perform_type: type[BasePerform]) -> None:
        ...

    def inject(self, perform_type: ..., *args, **kwargs):
        perform = perform_type(self)
        perform.__post_init__(*args, **kwargs)
        self.artifact_collections.insert(0, perform.__collector__.artifacts)

    async def maintain(self, perform: BasePerform):
        await self.exit_stack.enter_async_context(perform.lifespan())

    @asynccontextmanager
    async def lifespan(self):
        async with self.exit_stack:
            yield self

    def ext(self, components: dict[str, Any]):
        instance = copy(self)
        instance.components.update(components)
        return instance

    def get_fn_call(self, fn: Fn[P, R]) -> Callable[P, R]:
        def wrapper(*args: P.args, **kwargs: P.kwargs):
            return self.call_fn(fn, *args, **kwargs)
        return wrapper



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Callable, List, Optional, Union

from arclet.alconna import Alconna, CommandMeta
from arclet.alconna.args import Arg, Args, TAValue
from arclet.alconna.tools import AlconnaString
from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.saya.behaviour import Behaviour
from graia.saya.cube import Cube
from graia.saya.schema import BaseSchema
from pygtrie import _NoChildren

from . import AvillaCommands


class Command(AlconnaString, BaseSchema):
    def __init__(
        self,
        command: str,
        help_text: Optional[str] = None,
        dispatchers: Optional[List[BaseDispatcher]] = None,
        decorators: Optional[List[Decorator]] = None,
    ):
        super().__init__(command, help_text)
        self.dispatchers = dispatchers or []
        self.decorators = decorators or []

    def register(self, func: Callable, cmd: AvillaCommands):
        """注册 func 至 AvillaCommands

        Args:
            func (Callable): 命令函数
            cmd (AvillaCommands): 命令对象
        """
        cmd.on(self.build(), self.dispatchers, self.decorators)(func)  # type: ignore


@dataclass
class On(BaseSchema):
    """命令监听 Schema, 相当于 avilla_commands.on"""

    command: Union[str, Alconna]
    dispatchers: List[BaseDispatcher] = field(default_factory=list)
    decorators: List[Decorator] = field(default_factory=list)
    args: Optional[dict[str, Union[TAValue, Args, Arg]]] = field(default=None)
    meta: Optional[CommandMeta] = field(default=None)

    def register(self, func: Callable, cmd: AvillaCommands):
        """注册 func 至 AvillaCommands

        Args:
            func (Callable): 命令函数
            cmd (AvillaCommands): 命令对象
        """
        cmd.on(self.command, self.dispatchers, self.decorators, args=self.args, meta=self.meta)(func)  # type: ignore


class AvillaCommandsBehaviour(Behaviour):
    """命令行为"""

    def __init__(self, cmd: AvillaCommands) -> None:
        self.cmd = cmd

    def allocate(self, cube: Cube[Union[On, Command]]):
        if not isinstance(cube.metaclass, (On, Command)):
            return
        cube.metaclass.register(cube.content, self.cmd)
        return True

    def release(self, cube: Cube[Union[On, Command]]):
        if not isinstance(cube.metaclass, (On, Command)):
            return
        for key, target in self.cmd.trie.items():
            if TYPE_CHECKING:
                assert not isinstance(target, _NoChildren)

            if target[1].callable is cube.content:
                del self.cmd.trie[key]
                break

        return True

    uninstall = release



from __future__ import annotations

from functools import reduce
from typing import TYPE_CHECKING, Any, Callable, ChainMap, overload

from typing_extensions import ParamSpec, TypeVar, Unpack

from avilla.core.builtins.capability import CoreCapability
from avilla.core.metadata import MetadataRoute
from avilla.core.selector import (
    FollowsPredicater,
    Selector,
    _FollowItem,
    _parse_follows,
)
from graia.ryanvk import BaseCollector
from graia.ryanvk import Staff as BaseStaff

from .descriptor.query import find_querier_steps, query_depth_generator

if TYPE_CHECKING:
    from avilla.core.metadata import Metadata
    from avilla.core.resource import Resource

    from .descriptor.query import QueryHandler, QueryHandlerPerform


T = TypeVar("T")
R = TypeVar("R", covariant=True)
M = TypeVar("M", bound="Metadata")
P = ParamSpec("P")
P1 = ParamSpec("P1")
Co = TypeVar("Co", bound="BaseCollector")


class Staff(BaseStaff):
    """手杖与核心工艺 (Staff & Focus Craft)."""

    def get_context(self, target: Selector, *, via: Selector | None = None):
        return self.call_fn(CoreCapability.get_context, target, via=via)

    async def fetch_resource(self, resource: Resource[T]) -> T:
        return await self.get_fn_call(CoreCapability.fetch)(resource)

    @overload
    async def pull_metadata(
        self,
        target: Selector,
        route: type[M],
    ) -> M:
        ...

    @overload
    async def pull_metadata(
        self,
        target: Selector,
        route: MetadataRoute[Unpack[tuple[Any, ...]], T],
    ) -> T:
        ...

    async def pull_metadata(
        self,
        target: Selector,
        route: ...,
    ):
        return await self.call_fn(CoreCapability.pull, target, route)

    async def query_entities(self, pattern: str, **predicators: FollowsPredicater):
        items = _parse_follows(pattern, **predicators)
        artifact_map = ChainMap(*self.artifact_collections)
        steps = find_querier_steps(artifact_map, items)

        if steps is None:
            return

        def build_handler(artifact: tuple[BaseCollector, QueryHandlerPerform]) -> QueryHandler:
            async def handler(predicate: Callable[[str, str], bool] | str, previous: Selector | None = None):
                instance, entity = artifact
                async for i in entity(instance, predicate, previous):
                    yield i

            return handler

        def build_predicate(_steps: tuple[_FollowItem, ...]) -> Callable[[str, str], bool]:
            mapping = {i.name: i for i in _steps}

            def predicater(key: str, value: str) -> bool:
                if key not in mapping:
                    raise KeyError(f"expected existed key: {key}")
                item = mapping[key]
                if item.literal is not None:
                    return value == item.literal
                elif item.predicate is not None:
                    return item.predicate(value)
                return True

            return predicater

        handlers = []
        for follow_item, query_record in steps:
            handlers.append((follow_item, build_handler(artifact_map[query_record])))

        r = reduce(
            lambda previous, current: query_depth_generator(current[1], build_predicate(current[0]), previous),
            handlers,
            None,
        )
        if TYPE_CHECKING:
            assert r is not None

        async for i in r:
            yield i



from __future__ import annotations

from graia.amnesia.message import MessageChain

from avilla.core.ryanvk import Capability, Fn, TargetOverload
from avilla.core.selector import Selector

# MessageFetch => rs.pull(Message, target=...)


class MessageSend(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def send(self, target: Selector, message: MessageChain, *, reply: Selector | None = None) -> Selector:
        ...


class MessageRevoke(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def revoke(self, target: Selector) -> None:
        ...


class MessageEdit(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def edit(self, target: Selector, content: MessageChain) -> None:
        ...



from __future__ import annotations

from contextlib import AbstractContextManager
from typing import TYPE_CHECKING, Any, Callable, TypeVar

from ._runtime import GLOBAL_GALLERY
from .perform import BasePerform

if TYPE_CHECKING:
    from .typing import P, R, SupportsCollect

T = TypeVar("T")


class BaseCollector:
    artifacts: dict[Any, Any]
    collected_callbacks: list[Callable[[type[BasePerform]], Any]]

    namespace: str | None = None
    identify: str | None = None

    def __init__(self, artifacts: dict[Any, Any] | None = None) -> None:
        self.artifacts = artifacts or {}
        self.collected_callbacks = [self.__post_collected__]

    def __post_collected__(self, cls: type[BasePerform]):
        self.cls = cls

        if self.namespace is not None:
            ns: dict = GLOBAL_GALLERY.setdefault(self.namespace, {})
            locate: dict = ns.setdefault(self.identify or "_", {})
            locate.update(self.artifacts)

    @property
    def _(self):
        class LocalPerformTemplate(BasePerform, native=True):
            __collector__ = self

        return LocalPerformTemplate

    @classmethod
    def get_forks_name(cls) -> str | None:
        ...

    def collect(self, signature: Any, artifact: Any):
        self.artifacts[signature] = artifact

    def on_collected(self, func: Callable[[type], Any]):
        self.collected_callbacks.append(func)

    def forks(self, collector: BaseCollector):
        maybe_name = collector.get_forks_name()
        if maybe_name is None:
            raise TypeError(f"{type(collector)} doesn't support fork")
        collector.artifacts = self.artifacts.setdefault(maybe_name, {})
        collector.collected_callbacks = self.collected_callbacks
        return collector

    def attachs(self, collector: BaseCollector):
        collector.artifacts = self.artifacts
        collector.collected_callbacks = self.collected_callbacks
        return collector

    def using(self, context_manager: AbstractContextManager[T]) -> T:
        self.on_collected(lambda _: context_manager.__exit__(None, None, None))
        return context_manager.__enter__()

    def post_merge(self, origin: dict):
        self.on_collected(lambda _: origin.update(self.artifacts))

    def entity(self, signature: SupportsCollect[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
        return signature.collect(self, *args, **kwargs)



import importlib.metadata
import random
import re
from inspect import cleandoc

from loguru import logger

AVILLA_ASCII_LOGO = cleandoc(
    r"""
    [bold]Avilla[/]: a universal backend/framework for effective im development, powered by [blue]Graia Project[/].
        _        _ _ _
       / \__   _(_) | | __ _
      / _ \ \ / / | | |/ _` |
     / ___ \ V /| | | | (_| |
    /_/   \_\_/ |_|_|_|\__,_|
    """
)
AVILLA_ASCII_RAW_LOGO = re.sub(r"\[.*?\]", "", AVILLA_ASCII_LOGO)


if random.choice((True, False)):
    AVILLA_ASCII_LOGO += "\n"
    AVILLA_ASCII_LOGO += "[dim]when will I be able to find a place that can be my second hometown...?[/]"

    # but no matter what, the most important thing for me is always to be with my only loved one.
    # so that the journey will not be lonely.


GRAIA_PROJECT_REPOS = [
    "avilla-core",
    "graia-broadcast",
    "graia-saya",
    "graia-scheduler",
    "graia-ariadne",
    "statv",
    "launart",
    "creart",
    "creart-graia",
    "kayaku",
]


def log_telemetry():
    for telemetry in GRAIA_PROJECT_REPOS:
        try:
            version = importlib.metadata.version(telemetry)
        except Exception:
            version = "unknown / not-installed"
        logger.info(
            f"{telemetry} version: {version}",
            alt=f"[b cornflower_blue]{telemetry}[/] version: [cyan3]{version}[/]",
        )



from __future__ import annotations

from typing import Any, Callable

from avilla.core.metadata import Route
from graia.ryanvk.collector import BaseCollector
from graia.ryanvk.overload import FnOverload


class MetadataOverload(FnOverload):
    def collect_entity(
        self,
        collector: BaseCollector,
        scope: dict[Any, Any],
        entity: Any,
        params: dict[str, Route],
    ) -> None:
        record = (collector, entity)

        for param, route in params.items():
            ...
            collection: dict[Route, set] = scope.setdefault(param, {})
            collection.setdefault(route, set()).add(record)

    def get_entities(self, scope: dict[Any, Any], args: dict[str, Route]) -> set[tuple[BaseCollector, Callable]]:
        sets: list[set] = []

        for arg_name, route in args.items():
            if arg_name not in scope:
                raise NotImplementedError

            collection = scope[arg_name]

            if route not in collection:
                raise NotImplementedError

            sets.append(collection[route])

        return sets.pop().intersection(*sets)

    def merge_scopes(self, *scopes: dict[Any, Any]):
        # layout: {param_name: {route: set()}}
        result = {}

        for scope in scopes:
            for param_name, collection in scope.items():
                result_collection = result.setdefault(param_name, {})
                for route, entities in collection.items():
                    routes = result_collection.setdefault(route, set())
                    routes.update(entities)

        return result



from .capability import NickCapability as NickCapability
from .capability import SummaryCapability as SummaryCapability
from .metadata import Nick as Nick
from .metadata import Summary as Summary
from .metadata import Avatar as Avatar



from .capability import RequestCapability as RequestCapability
from .event import RequestAccepted as RequestAccepted
from .event import RequestCancelled as RequestCancelled
from .event import RequestEvent as RequestEvent
from .event import RequestIgnored as RequestIgnored
from .event import RequestReceived as RequestReceived
from .event import RequestRejected as RequestRejected
from .metadata import Answers as Answers
from .metadata import Comment as Comment
from .metadata import Questions as Questions
from .metadata import Reason as Reason
